# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import math
from collections import deque
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Callable

import numpy as np
from qm import qua
from qm.program import Program
from qm.qua import _dsl as qua_dsl
from qualang_tools.config.integration_weights_tools import convert_integration_weights

from qililab.qprogram.blocks import Average, Block, ForLoop, Loop, Parallel
from qililab.qprogram.blocks.infinite_loop import InfiniteLoop
from qililab.qprogram.calibration import Calibration
from qililab.qprogram.operations import (
    Measure,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.variable import Domain, FloatVariable, IntVariable, Variable
from qililab.waveforms import IQPair, Square, Waveform

# mypy: disable-error-code="operator"


@dataclass
class LoopInfo:
    """Class representing the output information generated by QbloxCompiler for an acquisition."""

    iterations: int
    is_average: bool


class _BusCompilationInfo:
    def __init__(self) -> None:
        self.current_gain: float | qua.QuaVariableType | None = None
        self.threshold: float | None = None
        self.threshold_rotation: float | None = None


class _MeasurementCompilationInfo:
    def __init__(
        self,
        bus: str,
        variable_I: qua.QuaVariableType,
        variable_Q: qua.QuaVariableType,
        stream_I: qua_dsl._ResultSource,
        stream_Q: qua_dsl._ResultSource,
        stream_raw_adc: qua_dsl._ResultSource | None = None,
    ) -> None:
        self.bus: str = bus
        self.variable_I: qua.QuaVariableType = variable_I
        self.variable_Q: qua.QuaVariableType = variable_Q
        self.stream_I: qua_dsl._ResultSource = stream_I
        self.stream_Q: qua_dsl._ResultSource = stream_Q
        self.stream_raw_adc: qua_dsl._ResultSource | None = stream_raw_adc
        self.loop_info_list: list[LoopInfo] = []


class MeasurementInfo:
    """Class representing information about the measurements taking place."""

    def __init__(self, bus: str, result_handles: list[str], threshold: float = 0.0):
        self.bus: str = bus
        self.result_handles: list[str] = result_handles
        self.threshold: float = threshold


class QuantumMachinesCompilationOutput:
    """Class representing the output information generated by QbloxCompiler.

    Attributes:
        qua (Program): The generated QUA program.
        config (dict): The generated configuration.
        measurements (list[MeasurementInfo]): List of measurement information.
    """

    def __init__(
        self, qprogram: QProgram, qua: Program, configuration: dict, measurements: list[MeasurementInfo]
    ) -> None:
        """Initialize the QuantumMachinesCompilationOutput instance.

        Args:
            qua (Program): The generated QUA program.
            config (dict): The configuration for the program.
            measurements (list[MeasurementInfo]): The list of measurement info objects.
        """
        self.qprogram = qprogram
        self.qua: Program = qua
        self.configuration: dict = configuration
        self.measurements: list[MeasurementInfo] = measurements

    def __iter__(self):
        """Allows the class to be unpacked as a tuple (program, config, measurements)."""
        yield self.qua
        yield self.configuration
        yield self.measurements


class QuantumMachinesCompiler:
    """A class for compiling QProgram to Quantum Machines hardware."""

    FREQUENCY_COEFF = 1
    PHASE_COEFF = 2 * np.pi
    VOLTAGE_COEFF = 2
    WAIT_COEFF = 4
    MINIMUM_TIME = 4

    def __init__(self) -> None:
        # Handlers to map each operation to a corresponding handler function
        self._handlers: dict[type, Callable] = {
            InfiniteLoop: self._handle_infinite_loop,
            Parallel: self._handle_parallel_loop,
            ForLoop: self._handle_for_loop,
            Loop: self._handle_loop,
            Average: self._handle_average,
            Measure: self._handle_measure,
            Play: self._handle_play,
            SetFrequency: self._handle_set_frequency,
            SetOffset: self._handle_set_offset,
            SetPhase: self._handle_set_phase,
            SetGain: self._handle_set_gain,
            ResetPhase: self._handle_reset_phase,
            Wait: self._handle_wait,
            Sync: self._handle_sync,
            Block: self._handle_block,
        }

        self._qprogram: QProgram
        self._qprogram_block_stack: deque[Block]
        self._qprogram_to_qua_variables: dict[Variable, qua.QuaVariableType]
        self._measurements: list[_MeasurementCompilationInfo]
        self._configuration: dict
        self._buses: dict[str, _BusCompilationInfo]

    def compile(
        self,
        qprogram: QProgram,
        bus_mapping: dict[str, str] | None = None,
        thresholds: dict[str, float] | None = None,
        threshold_rotations: dict[str, float] | None = None,
        calibration: Calibration | None = None,
    ) -> QuantumMachinesCompilationOutput:
        """Compile QProgram to QUA's Program.

        Args:
            qprogram (QProgram): The QProgram to be compiled
            bus_mapping (dict[str, str], optional): Optional mapping of bus names. Defaults to None.

        Returns:
            tuple[qua.Program, dict, list[MeasurementInfo]]: A tuple consisting of the generated QUA program, the generated configuration and a list of measurements' information.
        """

        def traverse(block: Block):
            self._qprogram_block_stack.append(block)
            for element in block.elements:
                handler = self._handlers.get(type(element))
                if not handler:
                    raise NotImplementedError(
                        f"{element.__class__} operation is currently not supported in Quantum Machines."
                    )
                if isinstance(element, (InfiniteLoop, ForLoop, Loop, Average, Parallel, Block)):
                    with handler(element):
                        traverse(element)
                else:
                    handler(element)
            self._qprogram_block_stack.pop()

        self._qprogram = qprogram
        if bus_mapping is not None:
            self._qprogram = self._qprogram.with_bus_mapping(bus_mapping=bus_mapping)
        if calibration is not None:
            self._qprogram = self._qprogram.with_calibration(calibration=calibration)
        if self._qprogram.has_calibrated_waveforms_or_weights():
            raise RuntimeError(
                "Cannot compile to hardware-native instructions because QProgram contains named operations that are not mapped. Provide a calibration instance containing all necessary mappings."
            )

        self._qprogram_block_stack = deque()
        self._qprogram_to_qua_variables = {}
        self._measurements = []
        self._configuration = {
            "waveforms": {},
            "pulses": {},
            "integration_weights": {},
            "digital_waveforms": {"ON": {"samples": [(1, 0)]}, "OFF": {"samples": [(0, 0)]}},
            "elements": {},
        }

        self._populate_buses()

        # Pre-processing: Update rotation threshold
        if threshold_rotations is not None:
            for bus in self._buses.keys() & threshold_rotations.keys():
                self._buses[bus].threshold_rotation = threshold_rotations[bus]

        if thresholds is not None:
            for bus in self._buses.keys() & thresholds.keys():
                self._buses[bus].threshold = thresholds[bus]

        with qua.program() as qua_program:
            # Declare variables
            self._declare_variables()
            # Recursive traversal to convert QProgram to QUA program.
            traverse(self._qprogram.body)
            # Stream Processing
            with qua.stream_processing():
                measurements = self._process_measurements()

        # Return a dictionary with bus names as keys and the compiled Sequence as values.
        return QuantumMachinesCompilationOutput(
            qprogram=self._qprogram, qua=qua_program, configuration=self._configuration, measurements=measurements
        )

    def _process_measurements(self):
        def _process_measurement(measurement: _MeasurementCompilationInfo, index: int):
            result_handles: list[str] = []

            def _process_stream(stream: qua_dsl._ResultSource, save_as: str) -> None:
                processing_stream: qua_dsl._ResultSource | qua_dsl._ResultStream = stream
                for index, loop_info in enumerate(measurement.loop_info_list):
                    if loop_info.is_average:
                        if index < len(measurement.loop_info_list) - 1:
                            processing_stream = processing_stream.buffer(loop_info.iterations)
                            processing_stream = processing_stream.map(["average"])
                        else:
                            processing_stream = processing_stream.average()
                    else:
                        processing_stream = processing_stream.buffer(loop_info.iterations)
                processing_stream.save(save_as)
                result_handles.append(save_as)

            _process_stream(measurement.stream_I, f"I_{index}")
            _process_stream(measurement.stream_Q, f"Q_{index}")
            if measurement.stream_raw_adc is not None:
                _process_stream(measurement.stream_raw_adc.input1(), f"adc1_{index}")
                _process_stream(measurement.stream_raw_adc.input2(), f"adc2_{index}")

            return result_handles

        measurements = [
            MeasurementInfo(
                bus=measurement.bus,
                result_handles=_process_measurement(measurement=measurement, index=i),
                threshold=self._buses[measurement.bus].threshold or 0.0,
            )
            for i, measurement in enumerate(self._measurements)
        ]

        return measurements

    def _declare_variables(self):
        for variable in self._qprogram.variables:
            if variable.domain in [Domain.Time, Domain.Frequency] or (
                variable.domain is Domain.Scalar and isinstance(variable, IntVariable)
            ):
                qua_variable = qua.declare(int)
            else:
                qua_variable = qua.declare(qua.fixed)
            self._qprogram_to_qua_variables[variable] = qua_variable

    def _populate_buses(self):
        """Map each bus in the QProgram to a BusCompilationInfo instance."""

        buses = self._qprogram.buses
        self._configuration["elements"] = {bus: {"operations": {}} for bus in buses}
        self._buses = {bus: _BusCompilationInfo() for bus in buses}

    def _handle_infinite_loop(self, _: InfiniteLoop):
        return qua.infinite_loop_()

    def _handle_parallel_loop(self, element: Parallel):
        variables = []
        loops = []
        for loop in element.loops:
            qua_variable = self._qprogram_to_qua_variables[loop.variable]
            values = (
                np.arange(loop.start, loop.stop + loop.step / 2, loop.step)
                if isinstance(loop, ForLoop)
                else loop.values
            )
            if loop.variable.domain is Domain.Phase:
                values = values / self.PHASE_COEFF
            if loop.variable.domain is Domain.Frequency:
                values = values.astype(int)
            if loop.variable.domain is Domain.Scalar and isinstance(loop.variable, IntVariable):
                values = values.astype(int)
            if loop.variable.domain is Domain.Time:
                values = np.maximum(values, self.MINIMUM_TIME).astype(int)

            variables.append(qua_variable)
            loops.append(values)
        return qua.for_each_(tuple(variables), tuple(loops))

    def _handle_for_loop(self, element: ForLoop):
        qua_variable = self._qprogram_to_qua_variables[element.variable]
        start, stop, step = element.start, element.stop, element.step

        if isinstance(element.variable, FloatVariable):
            stop += step / 2
        if element.variable.domain is Domain.Phase:
            start, stop, step = start / self.PHASE_COEFF, stop / self.PHASE_COEFF, step / self.PHASE_COEFF
        if element.variable.domain is Domain.Frequency:
            start, stop, step = (
                int(start),
                int(stop),
                int(step),
            )
        if element.variable.domain is Domain.Time:
            start = max(start, self.MINIMUM_TIME)

        to_positive = stop >= start
        if to_positive:
            return qua.for_(qua_variable, start, qua_variable <= stop, qua_variable + step)  # type: ignore[arg-type]
        return qua.for_(qua_variable, start, qua_variable >= stop, qua_variable + step)  # type: ignore[arg-type]

    def _handle_loop(self, element: Loop):
        qua_variable = self._qprogram_to_qua_variables[element.variable]
        values = element.values
        if element.variable.domain is Domain.Phase:
            values = values / self.PHASE_COEFF
        if element.variable.domain is Domain.Frequency:
            values = values.astype(int)
        if element.variable.domain is Domain.Scalar and isinstance(element.variable, IntVariable):
            values = values.astype(int)
        if element.variable.domain is Domain.Time:
            values = np.maximum(values, self.MINIMUM_TIME).astype(int)
        return qua.for_each_(qua_variable, values.tolist())

    def _handle_average(self, element: Average):
        variable = qua.declare(int)
        return qua.for_(variable, 0, variable < element.shots, variable + 1)  # type: ignore[arg-type]

    def _handle_set_frequency(self, element: SetFrequency):
        frequency = (
            self._qprogram_to_qua_variables[element.frequency]
            if isinstance(element.frequency, Variable)
            else element.frequency
        )
        qua.update_frequency(element=element.bus, new_frequency=frequency)

    def _handle_set_offset(self, element: SetOffset):
        if element.offset_path1 or isinstance(element.offset_path1, Variable):
            offset_i = (
                self._qprogram_to_qua_variables[element.offset_path0]
                if isinstance(element.offset_path0, Variable)
                else element.offset_path0
            )
            offset_q = (
                self._qprogram_to_qua_variables[element.offset_path1]
                if isinstance(element.offset_path1, Variable)
                else element.offset_path1
            )
            qua.set_dc_offset(element=element.bus, element_input="I", offset=offset_i)
            qua.set_dc_offset(element=element.bus, element_input="Q", offset=offset_q)
        else:
            offset = (
                self._qprogram_to_qua_variables[element.offset_path0]
                if isinstance(element.offset_path0, Variable)
                else element.offset_path0
            )
            qua.set_dc_offset(element=element.bus, element_input="single", offset=offset)

    def _handle_set_phase(self, element: SetPhase):
        phase = (
            self._qprogram_to_qua_variables[element.phase]
            if isinstance(element.phase, Variable)
            else element.phase / self.PHASE_COEFF
        )
        qua.frame_rotation_2pi(phase, element.bus)

    def _handle_reset_phase(self, element: ResetPhase):
        qua.reset_frame(element.bus)

    def _handle_set_gain(self, element: SetGain):
        gain = self._qprogram_to_qua_variables[element.gain] if isinstance(element.gain, Variable) else element.gain
        # QUA doesn't have a method for setting the gain directly.
        # Instead, it uses an amplitude multiplication with `amp()`.
        # Thus, we store the current gain to multiply with in the next play instructions.
        self._buses[element.bus].current_gain = gain

    def _handle_play(self, element: Play):
        waveform_I, waveform_Q = element.get_waveforms()
        waveform_variables = element.get_waveform_variables()
        duration = waveform_I.get_duration()

        gain = (
            qua.amp(self._buses[element.bus].current_gain * self.VOLTAGE_COEFF)  # type: ignore[arg-type]
            if self._buses[element.bus].current_gain is not None
            else None
        )

        if not waveform_variables:
            waveform_I_name = self.__add_waveform_to_configuration(waveform_I)
            waveform_Q_name = self.__add_waveform_to_configuration(waveform_Q) if waveform_Q else None
            pulse_name = self.__add_or_update_control_pulse_to_configuration(waveform_I_name, waveform_Q_name, duration)
            operation_name = self.__add_pulse_to_element_operations(element.bus, pulse_name)
            pulse = operation_name * gain if gain is not None else operation_name
            qua.play(pulse, element.bus)

    @contextmanager
    def _handle_block(self, element: Block):
        yield

    def _handle_measure(self, element: Measure):
        waveform_I, waveform_Q = element.get_waveforms()

        waveform_I_name = self.__add_waveform_to_configuration(waveform_I)
        waveform_Q_name = self.__add_waveform_to_configuration(waveform_Q)

        gain = (
            qua.amp(self._buses[element.bus].current_gain * self.VOLTAGE_COEFF)  # type: ignore[arg-type]
            if self._buses[element.bus].current_gain is not None
            else None
        )
        rotation = (
            element.rotation  # type: ignore
            if element.rotation is not None
            else (
                self._buses[element.bus].threshold_rotation
                if self._buses[element.bus] and self._buses[element.bus].threshold_rotation is not None
                else 0.0
            )
        )

        variable_I = qua.declare(qua.fixed)
        variable_Q = qua.declare(qua.fixed)
        stream_I = qua.declare_stream()
        stream_Q = qua.declare_stream()
        stream_raw_adc = qua.declare_stream(adc_trace=True) if element.save_adc else None

        A, B, C, D = self.__add_weights_to_configuration(weights=element.weights, rotation=rotation)  # type: ignore

        pulse_name = self.__add_or_update_measurement_pulse_to_configuration(
            waveform_I_name,
            waveform_Q_name,
            duration=waveform_I.get_duration(),
            integration_weights=[A, B, C, D],
        )
        operation_name = self.__add_pulse_to_element_operations(element.bus, pulse_name)
        pulse = operation_name * gain if gain is not None else operation_name
        if element.demodulation:
            qua.measure(
                pulse,
                element.bus,
                stream_raw_adc,
                qua.dual_demod.full(A, "out1", B, "out2", variable_I),
                qua.dual_demod.full(C, "out1", D, "out2", variable_Q),
            )
        else:
            qua.measure(
                pulse,
                element.bus,
                stream_raw_adc,
                qua.dual_integration.full(A, "out1", B, "out2", variable_I),
                qua.dual_integration.full(C, "out1", D, "out2", variable_Q),
            )
        qua.save(variable_I, stream_I)
        qua.save(variable_Q, stream_Q)

        measurement_info = _MeasurementCompilationInfo(
            element.bus, variable_I, variable_Q, stream_I, stream_Q, stream_raw_adc
        )
        for block in reversed(self._qprogram_block_stack):
            if isinstance(block, ForLoop):
                iterations = QuantumMachinesCompiler._calculate_iterations(block.start, block.stop, block.step)
                measurement_info.loop_info_list.append(LoopInfo(is_average=False, iterations=iterations))
            if isinstance(block, Loop):
                iterations = len(block.values)
                measurement_info.loop_info_list.append(LoopInfo(is_average=False, iterations=iterations))
            if isinstance(block, Parallel):
                iterations = min(
                    (
                        len(loop.values)
                        if isinstance(loop, Loop)
                        else QuantumMachinesCompiler._calculate_iterations(loop.start, loop.stop, loop.step)
                    )
                    for loop in block.loops
                )
                measurement_info.loop_info_list.append(LoopInfo(is_average=False, iterations=iterations))
            if isinstance(block, Average):
                measurement_info.loop_info_list.append(LoopInfo(is_average=True, iterations=block.shots))

        self._measurements.append(measurement_info)

    def _handle_wait(self, element: Wait):
        duration = (
            self._qprogram_to_qua_variables[element.duration]
            if isinstance(element.duration, Variable)
            else max(element.duration, self.MINIMUM_TIME)
        )

        qua.wait(
            (
                duration / int(self.WAIT_COEFF)  # type: ignore[arg-type]
                if isinstance(element.duration, Variable)
                else int(duration / self.WAIT_COEFF)  # type: ignore[arg-type]
            ),
            element.bus,
        )

    def _handle_sync(self, element: Sync):
        if element.buses:
            qua.align(*element.buses)
        else:
            qua.align()

    def __add_pulse_to_element_operations(self, element: str, pulse: str):
        if pulse not in self._configuration["elements"][element]["operations"]:
            self._configuration["elements"][element]["operations"][pulse] = pulse
        return pulse

    def __add_or_update_control_pulse_to_configuration(
        self, waveform_I_name: str, waveform_Q_name: str | None, duration: int
    ):
        pulse_name = (
            f"control_{waveform_I_name}_{waveform_Q_name}_{duration}"
            if waveform_Q_name
            else f"control_{waveform_I_name}_{duration}"
        )
        if pulse_name not in self._configuration["pulses"]:
            configuration = {
                "operation": "control",
                "length": duration,
                "waveforms": (
                    {"I": waveform_I_name, "Q": waveform_Q_name} if waveform_Q_name else {"single": waveform_I_name}
                ),
            }
            self._configuration["pulses"][pulse_name] = configuration
        return pulse_name

    def __add_or_update_measurement_pulse_to_configuration(
        self, waveform_I_name: str, waveform_Q_name: str, duration: int, integration_weights: list[str]
    ):
        pulse_name = f"measurement_{waveform_I_name}_{waveform_Q_name}_{duration}"
        if pulse_name not in self._configuration["pulses"]:
            self._configuration["pulses"][pulse_name] = {
                "operation": "measurement",
                "length": duration,
                "waveforms": {
                    "I": waveform_I_name,
                    "Q": waveform_Q_name,
                },
                "integration_weights": {weight: weight for weight in integration_weights},
                "digital_marker": "ON",
            }
        else:
            self._configuration["pulses"][pulse_name]["integration_weights"].update(
                {weight: weight for weight in integration_weights}
            )
        return pulse_name

    def __add_weights_to_configuration(self, weights: IQPair, rotation: float):
        prefix = f"{QuantumMachinesCompiler.__hash_waveform(weights.I)}_{QuantumMachinesCompiler.__hash_waveform(weights.Q)}_{rotation}"

        envelope_I = weights.I.envelope(4)
        envelope_Q = weights.Q.envelope(4)

        # Define weights as numpy array
        cos_I = np.cos(rotation) * envelope_I
        sin_I = np.sin(rotation) * envelope_I
        cos_Q = np.cos(rotation) * envelope_Q
        sin_Q = np.sin(rotation) * envelope_Q
        minus_sin_I = -np.sin(rotation) * envelope_I
        minus_cos_Q = -np.cos(rotation) * envelope_Q

        # Convert weights to QM-specific format
        cos_I_converted = convert_integration_weights(integration_weights=cos_I, N=len(cos_I))
        sin_I_converted = convert_integration_weights(integration_weights=sin_I, N=len(sin_I))
        cos_Q_converted = convert_integration_weights(integration_weights=cos_Q, N=len(cos_Q))
        sin_Q_converted = convert_integration_weights(integration_weights=sin_Q, N=len(sin_Q))
        minus_sin_I_converted = convert_integration_weights(integration_weights=minus_sin_I, N=len(minus_sin_I))
        minus_cos_Q_converted = convert_integration_weights(integration_weights=minus_cos_Q, N=len(minus_cos_Q))

        # Define weights names
        A = f"{prefix}_A"
        B = f"{prefix}_B"
        C = f"{prefix}_C"
        D = f"{prefix}_D"

        # Add weights to configuration dictionary
        self._configuration["integration_weights"][A] = {"cosine": cos_I_converted, "sine": sin_I_converted}
        self._configuration["integration_weights"][B] = {"cosine": minus_sin_I_converted, "sine": cos_I_converted}
        self._configuration["integration_weights"][C] = {"cosine": sin_Q_converted, "sine": minus_cos_Q_converted}
        self._configuration["integration_weights"][D] = {"cosine": cos_Q_converted, "sine": sin_Q_converted}

        # Return weights names
        return A, B, C, D

    def __add_waveform_to_configuration(self, waveform: Waveform):
        waveform_name = QuantumMachinesCompiler.__hash_waveform(waveform)
        if waveform_name not in self._configuration["waveforms"]:
            self._configuration["waveforms"][waveform_name] = QuantumMachinesCompiler.__waveform_to_config(waveform)
        return waveform_name

    @staticmethod
    def __hash_waveform(waveform: Waveform):
        attributes = [
            f"{key}: {(QuantumMachinesCompiler.__hash_waveform(value) if isinstance(value, Waveform) else value.tobytes() if isinstance(value, np.ndarray) else str(value))}"  # type: ignore[str-bytes-safe]
            for key, value in waveform.__dict__.items()
            if key != "duration" or not isinstance(waveform, Square)
        ]
        string_to_hash = f"{waveform.__class__.__name__}({','.join(attributes)})"
        hash_result = hashlib.md5(string_to_hash.encode("utf-8"), usedforsecurity=False)
        return hash_result.hexdigest()[:8]

    @staticmethod
    def __waveform_to_config(waveform: Waveform):
        if isinstance(waveform, Square):
            amplitude = waveform.amplitude / QuantumMachinesCompiler.VOLTAGE_COEFF
            return {"type": "constant", "sample": amplitude}

        envelope = waveform.envelope() / QuantumMachinesCompiler.VOLTAGE_COEFF
        return {"type": "arbitrary", "samples": envelope.tolist()}

    @staticmethod
    def _calculate_iterations(start: int | float, stop: int | float, step: int | float):
        if step == 0:
            raise ValueError("Step value cannot be zero")

        # Calculate the raw number of iterations
        raw_iterations = (stop - start + step) / step

        # If the raw number of iterations is very close to an integer, round it to that integer
        # This accounts for potential floating-point inaccuracies
        if abs(raw_iterations - round(raw_iterations)) < 1e-9:
            return round(raw_iterations)

        # Otherwise, if we're incrementing, take the ceiling, and if we're decrementing, take the floor
        return math.floor(raw_iterations) if step > 0 else math.ceil(raw_iterations)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass, field

from qililab.settings.analog.flux_control_topology import FluxControlTopology


@dataclass
class AnalogCompilationSettings:
    """Dataclass with all the settings and gates definitions needed to decompose gates into pulses."""

    flux_control_topology: list[FluxControlTopology] = field(default_factory=list)

    def __post_init__(self):
        """Build the Gates Settings based on the master settings."""
        self.flux_control_topology = [
            FluxControlTopology(**flux_control) if isinstance(flux_control, dict) else flux_control
            for flux_control in self.flux_control_topology
        ]

    def to_dict(self):
        """Serializes gate settings to dictionary and removes fields with None values"""

        return asdict(self)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""config.py"""

import logging
import os

from qililab import __version__

# Logging level from 0 (all) to 4 (errors) (see https://docs.python.org/3/library/logging.html#logging-levels)
LIBRARY_LOG_LEVEL = int(os.environ.get("LIBRARY_LOG_LEVEL", 20))


class CustomHandler(logging.StreamHandler):
    """Custom handler for logging algorithm."""

    def format(self, record):
        """Format the record with specific format."""

        fmt = f"[qililab] [{__version__}|%(levelname)s|%(asctime)s]: %(message)s"
        return logging.Formatter(fmt, datefmt="%Y-%m-%d %H:%M:%S").format(record)


logger = logging.getLogger(__name__)
logger.setLevel(LIBRARY_LOG_LEVEL)
logger.addHandler(CustomHandler())



"""Tests for the Qblox Module class."""
from unittest.mock import MagicMock

import pytest
from qpysequence import Acquisitions, Program, Sequence, Waveforms, Weights

from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.qblox import QbloxQCM
from qililab.platform import Platform
from qililab.data_management import build_platform
from qililab.typings import Parameter
from typing import cast
from qblox_instruments.qcodes_drivers.sequencer import Sequencer
from qblox_instruments.qcodes_drivers.module import Module as QcmQrm


@pytest.fixture(name="platform")
def fixture_platform():
    """platform fixture"""
    return build_platform(runcard="tests/instruments/qblox/qblox_runcard.yaml")


@pytest.fixture(name="qcm")
def fixture_qrm(platform: Platform):
    qcm = cast(QbloxQCM, platform.get_element(alias="qcm"))

    sequencer_mock_spec = [
        *Sequencer._get_required_parent_attr_names(),
        "sync_en",
        "gain_awg_path0",
        "gain_awg_path1",
        "sequence",
        "mod_en_awg",
        "nco_freq",
        "scope_acq_sequencer_select",
        "channel_map_path0_out0_en",
        "channel_map_path1_out1_en",
        "demod_en_acq",
        "integration_length_acq",
        "mixer_corr_phase_offset_degree",
        "mixer_corr_gain_ratio",
        "connect_out0",
        "connect_out1",
        "connect_out2",
        "connect_out3",
        "marker_ovr_en",
        "marker_ovr_value",
        "offset_awg_path0",
        "offset_awg_path1"
    ]

    module_mock_spec = [
        *QcmQrm._get_required_parent_qtm_attr_names(),
        "reference_source",
        "sequencer0",
        "sequencer1",
        "out0_offset",
        "out1_offset",
        "out2_offset",
        "out3_offset",
        "scope_acq_avg_mode_en_path0",
        "scope_acq_avg_mode_en_path1",
        "scope_acq_trigger_mode_path0",
        "scope_acq_trigger_mode_path1",
        "sequencers",
        "scope_acq_sequencer_select",
        "get_acquisitions",
        "disconnect_outputs",
        "disconnect_inputs",
        "arm_sequencer",
        "start_sequencer",
        "reset",
        "module_type"
    ]

    # Create a mock device using create_autospec to follow the interface of the expected device
    qcm.device = MagicMock()
    qcm.device.mock_add_spec(module_mock_spec)

    qcm.device.sequencers = {
        0: MagicMock(),
        1: MagicMock(),
    }

    for sequencer in qcm.device.sequencers:
        qcm.device.sequencers[sequencer].mock_add_spec(sequencer_mock_spec)

    return qcm


class TestQbloxQCM:
    def test_init(self, qcm: QbloxQCM):
        assert qcm.alias == "qcm"
        assert qcm.is_awg()
        assert not qcm.is_adc()
        assert len(qcm.awg_sequencers) == 2  # As per the YAML config
        assert qcm.out_offsets == [0.0, 0.1, 0.2, 0.3]
        sequencer = qcm.get_sequencer(0)
        assert sequencer.identifier == 0
        assert sequencer.outputs == [3, 2]
        assert sequencer.intermediate_frequency == 100e6
        assert sequencer.gain_imbalance == 0.05
        assert sequencer.phase_imbalance == 0.02
        assert sequencer.hardware_modulation is True
        assert sequencer.gain_i == 1.0
        assert sequencer.gain_q == 1.0
        assert sequencer.offset_i == 0.0
        assert sequencer.offset_q == 0.0

    def test_init_raises_error(self):
        with pytest.raises(ValueError):
            _ = build_platform(runcard="tests/instruments/qblox/qblox_qcm_too_many_sequencers_runcard.yaml")

    def test_module_type(self, qcm: QbloxQCM):
        _ = qcm.module_type()
        qcm.device.module_type.assert_called_once()

    @pytest.mark.parametrize(
        "parameter, value",
        [
            # Test GAIN setting
            (Parameter.GAIN, 2.0),
            (Parameter.GAIN, 3.5),

            # Test GAIN_I and GAIN_Q settings
            (Parameter.GAIN_I, 1.5),
            (Parameter.GAIN_Q, 1.5),

            # Test OFFSET_I and OFFSET_Q settings
            (Parameter.OFFSET_I, 0.1),
            (Parameter.OFFSET_Q, 0.2),

            # Test IF setting (intermediate frequency)
            (Parameter.IF, 100e6),

            # Test HARDWARE_MODULATION setting
            (Parameter.HARDWARE_MODULATION, True),

            # Test GAIN_IMBALANCE setting
            (Parameter.GAIN_IMBALANCE, 0.05),

            # Test PHASE_IMBALANCE setting
            (Parameter.PHASE_IMBALANCE, 0.02),

            # Test OFFSET_OUT settings
            (Parameter.OFFSET_OUT0, 0.1),
            (Parameter.OFFSET_OUT1, 0.15),
            (Parameter.OFFSET_OUT2, 0.2),
            (Parameter.OFFSET_OUT3, 0.25),
        ]
    )
    def test_set_parameter(self, qcm: QbloxQCM, parameter, value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        qcm.set_parameter(parameter, value, channel_id=0)
        sequencer = qcm.get_sequencer(0)

        # Check values based on the parameter
        if parameter == Parameter.GAIN:
            assert sequencer.gain_i == value
            assert sequencer.gain_q == value
        elif parameter == Parameter.GAIN_I:
            assert sequencer.gain_i == value
        elif parameter == Parameter.GAIN_Q:
            assert sequencer.gain_q == value
        elif parameter == Parameter.OFFSET_I:
            assert sequencer.offset_i == value
        elif parameter == Parameter.OFFSET_Q:
            assert sequencer.offset_q == value
        elif parameter == Parameter.IF:
            assert sequencer.intermediate_frequency == value
        elif parameter == Parameter.HARDWARE_MODULATION:
            assert sequencer.hardware_modulation == value
        elif parameter == Parameter.GAIN_IMBALANCE:
            assert sequencer.gain_imbalance == value
        elif parameter == Parameter.PHASE_IMBALANCE:
            assert sequencer.phase_imbalance == value
        elif parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            assert qcm.out_offsets[output] == value

    def test_set_parameter_gain(self, qcm: QbloxQCM):
        """Test handling invalid channel IDs when setting parameters."""
        qcm.set_parameter(Parameter.GAIN, 2.0, channel_id=0)
        sequencer = qcm.get_sequencer(0)
        assert sequencer.gain_i == 2.0
        assert sequencer.gain_q == 2.0

    def test_set_parameter_raises_error(self, qcm: QbloxQCM):
        """Test setting parameters for QCM sequencers using parameterized values."""
        with pytest.raises(ParameterNotFound):
            qcm.set_parameter(Parameter.BUS_FREQUENCY, value=42, channel_id=0)

        with pytest.raises(IndexError):
            qcm.set_parameter(Parameter.PHASE_IMBALANCE, value=0.5, channel_id=4)

        with pytest.raises(Exception):
            qcm.set_parameter(Parameter.PHASE_IMBALANCE, value=0.5, channel_id=None)

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            # Test GAIN_I and GAIN_Q settings
            (Parameter.GAIN_I, 1.0),
            (Parameter.GAIN_Q, 1.0),

            # Test OFFSET_I and OFFSET_Q settings
            (Parameter.OFFSET_I, 0.0),
            (Parameter.OFFSET_Q, 0.0),

            # Test IF setting (intermediate frequency)
            (Parameter.IF, 100e6),

            # Test HARDWARE_MODULATION setting
            (Parameter.HARDWARE_MODULATION, True),

            # Test GAIN_IMBALANCE setting
            (Parameter.GAIN_IMBALANCE, 0.05),

            # Test PHASE_IMBALANCE setting
            (Parameter.PHASE_IMBALANCE, 0.02),

            # Test OFFSET_OUT settings
            (Parameter.OFFSET_OUT0, 0.0),
            (Parameter.OFFSET_OUT1, 0.1),
            (Parameter.OFFSET_OUT2, 0.2),
            (Parameter.OFFSET_OUT3, 0.3),
        ]
    )
    def test_get_parameter(self, qcm: QbloxQCM, parameter, expected_value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        value = qcm.get_parameter(parameter, channel_id=0)
        assert value == expected_value

    def test_get_parameter_raises_error(self, qcm: QbloxQCM):
        """Test setting parameters for QCM sequencers using parameterized values."""
        with pytest.raises(ParameterNotFound):
            qcm.get_parameter(Parameter.BUS_FREQUENCY, channel_id=0)

        with pytest.raises(IndexError):
            qcm.get_parameter(Parameter.PHASE_IMBALANCE, channel_id=4)

        with pytest.raises(Exception):
            qcm.get_parameter(Parameter.PHASE_IMBALANCE, channel_id=None)

    def test_set_markers_override_enabled(self, qcm: QbloxQCM):
        qcm.set_markers_override_enabled(value=True, sequencer_id=0)
        qcm.device.sequencers[0].marker_ovr_en.assert_called_with(True)

        qcm.set_markers_override_enabled(value=False, sequencer_id=0)
        qcm.device.sequencers[0].marker_ovr_en.assert_called_with(False)

    def test_set_markers_override_value(self, qcm: QbloxQCM):
        qcm.set_markers_override_value(value=123, sequencer_id=0)
        qcm.device.sequencers[0].marker_ovr_value.assert_called_once_with(123)

    def test_initial_setup(self, qcm: QbloxQCM):
        """Test the initial setup of the QCM module."""
        qcm.initial_setup()

        # Verify the correct setup calls were made on the device
        qcm.device.disconnect_outputs.assert_called_once()
        for sequencer in qcm.awg_sequencers:
            qcm.device.sequencers[sequencer.identifier].sync_en.assert_called_with(False)

    def test_run(self, qcm: QbloxQCM):
        """Test running the QCM module."""
        qcm.sequences[0] = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights())
        qcm.run(channel_id=0)

        sequencer = qcm.get_sequencer(0)
        qcm.device.arm_sequencer.assert_called_with(sequencer=sequencer.identifier)
        qcm.device.start_sequencer.assert_called_with(sequencer=sequencer.identifier)

    def test_upload(self, qcm: QbloxQCM):
        """Test uploading a QpySequence to the QCM module."""
        qcm.sequences[0] = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights())
        qcm.upload(channel_id=0)

        qcm.device.sequencers[0].sequence.assert_called_once_with(qcm.sequences[0].todict())
        qcm.device.sequencers[0].sync_en.assert_called_once_with(True)

    def test_upload_qpysequence(self, qcm: QbloxQCM):
        """Test uploading a QpySequence to the QCM module."""
        sequence = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights())
        qcm.upload_qpysequence(qpysequence=sequence, channel_id=0)

        qcm.device.sequencers[0].sequence.assert_called_once_with(sequence.todict())

    def test_clear_cache(self, qcm: QbloxQCM):
        """Test clearing the cache of the QCM module."""
        qcm.cache = {0: MagicMock()}  # type: ignore[misc]
        qcm.clear_cache()

        assert qcm.cache == {}
        assert qcm.sequences == {}

    def test_reset(self, qcm: QbloxQCM):
        """Test resetting the QCM module."""
        qcm.reset()

        qcm.device.reset.assert_called_once()
        assert qcm.cache == {}
        assert qcm.sequences == {}

    def test_turn_off(self, qcm: QbloxQCM):
        qcm.turn_off()

        assert qcm.device.stop_sequencer.call_count == 2

    def test_sync_sequencer(self, qcm: QbloxQCM):
        qcm.sync_sequencer(sequencer_id=0)

        qcm.device.sequencers[0].sync_en.assert_called_once_with(True)

    def test_desync_sequencer(self, qcm: QbloxQCM):
        qcm.desync_sequencer(sequencer_id=0)

        qcm.device.sequencers[0].sync_en.assert_called_once_with(False)

    def test_desync_sequencers(self, qcm: QbloxQCM):
        qcm.desync_sequencers()

        for sequencer in qcm.awg_sequencers:
            qcm.device.sequencers[sequencer.identifier].sync_en.assert_called_once_with(False)

class TestQbloxSequencer:
    def test_to_dict(self, qcm: QbloxQCM):
        sequencer = qcm.get_sequencer(0)
        as_dict = sequencer.to_dict()
        assert isinstance(as_dict, dict)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Automatic-calibration Controller module, which works with notebooks as nodes."""

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import networkx as nx
import pandas as pd

from qililab import Parameter
from qililab.calibration.calibration_node import CalibrationNode
from qililab.config import logger
from qililab.data_management import build_platform, save_platform

if TYPE_CHECKING:
    from qililab.platform.platform import Platform


class CalibrationController:
    """Controls the automatic calibration sequence.

    **Usage:**
        - To calibrate the full graph, use the ``run_automatic_calibration()`` method.

    |

    **Graph Structure:**

    In the graph, directions should be given by `nodes` pointing to their next `dependents` (natural time flow for calibration).
    This defines our `starts` and `ends` of the calibration:

    - `starts`: `Roots` (``in_degree=0``, no arrows pointing in, no `dependencies`) of the graph, where all arrows leave the node.

    - `ends`: `Leaves` (``out_degree=0``, no arrows pointing out, no `dependants`) of the graph, where all arrows enter the node.

    .. code-block:: python

        #                     /--> 2 -->\\
        #                    /     |     \\
        #    (start, root)  0      |      3 --> 4 (end, leave)
        #                    \\     v     /
        #                     \\--> 1 -->/

    .. note::

        Find information about the automatic-calibration workflow, in the examples below.

    Args:
        calibration_graph (nx.DiGraph): The calibration (directed acyclic) graph, where each node is a ``string`` corresponding to a ``CalibrationNode.node_id``. Directions should be given
            by `nodes` pointing to their next `dependents` (natural time flow for calibration), defining our `starts` and `ends` of the calibration as the `roots` (``in_degree=0``) and `leaves`
            (``out_degree=0``) of the graph.
        node_sequence (dict[str, CalibrationNode]): Mapping for the nodes of the graph, from strings into the actual initialized nodes.
        runcard (str): The runcard path, containing the serialized platform where the experiments will be run.
        drift_timeout (float, optional): Duration in seconds, representing an estimate of how long it takes for the calibration parameters to drift.
            During that time the parameters of this node should be considered calibrated. Thus a big value will tend to skip recently calibrated nodes,
            making the calibration process faster, but less accurate,and a small value will make the calibration process slower, but more accurate and robust.
            A node will be skipped if the ``drift timeout`` is bigger than the time since its last calibration. Defaults to 7200 (3h).

    Examples:

        **Calibration Workflow:**

        The calibration process is structured into three levels of methods:

        1. **Highest Level Method**: The ``run_automatic_calibration()`` method finds all the end nodes of the graph (`leaves`, those without further `dependents`) and runs ``calibrate_all()`` on them.

        2. **Mid-Level Method**: ``calibrate_all()``.
            - ``calibrate_all(node)`` starts from the `roots` that ``node`` depends on, and moves forwards (`dependency -> dependant`) until ``node``, checking the last time executions at each step.

        3. **Low-Level Method**: ``calibrate()`` is the method you would be calling during this process to interact with the ``nodes``.


        ----------

        **Dangerous Behaviors:**

        Note that depending on your ``CalibrationController`` construction, you can have dangerous behaviors in the workflow. You need to watch out for:
        - If you give too long ``drift_timeout``'s, since ``calibrate_all()`` will assume the node is 100% working.. To start the calibration from the start again, just reduce the ``drift_timeout``, or remove the executed files!

        ----------

        **Practical example:**

        To create two linked nodes twice, for two different qubits, and pass them to a :class:`.CalibrationController` and run a ``calibrate_all()``, you need:

        .. code-block:: python

            import numpy as np
            import networkx as nx

            from qililab.calibration import CalibrationController, CalibrationNode, norm_root_mean_sqrt_error

            # GRAPH CREATION AND NODE MAPPING (key = name in graph, value = node object):
            nodes = {}
            G = nx.DiGraph()
            first, second = [], []

            # CREATE NODES:
            for qubit in [0, 1]:
                first[qubit] = CalibrationNode(
                    nb_path="notebooks/first.ipynb",
                    qubit_index=qubit,
                )
                nodes[first[qubit].node_id] = first[qubit]

                second[qubit] = CalibrationNode(
                    nb_path="notebooks/second.ipynb",
                    qubit_index=qubit,
                    sweep_interval=np.arange(start=0, stop=19, step=1),
                )
                nodes[second[qubit].node_id] = second[qubit]

                # GRAPH BUILDING (1 --> 2):
                G.add_edge(first[qubit].node_id, second[qubit].node_id)

            # CREATE CALIBRATION CONTROLLER:
            controller = CalibrationController(node_sequence=nodes, calibration_graph=G, runcard=path_runcard)

            ### WORKFLOW TO DO:
            controller.calibrate_all(second[1]) # calibrate_all starting at second node for qubit 1

        .. note::

            Find information about how these nodes and their notebooks need to be in the :class:`CalibrationNode` class documentation.

            There you will also find the above code, but without defining ``first`` and ``second`` as lists.
    """

    def __init__(
        self,
        calibration_graph: nx.DiGraph,
        node_sequence: dict[str, CalibrationNode],
        runcard: str,
        drift_timeout: float = 7200.0,
    ):
        if not nx.is_directed_acyclic_graph(calibration_graph):
            raise ValueError("The calibration graph must be a Directed Acyclic Graph (DAG).")

        self.calibration_graph: nx.DiGraph = calibration_graph
        """The calibration (directed acyclic) graph. Where each node is a ``string`` corresponding to a ``CalibrationNode.node_id``.

        Directions should be given by `nodes` pointing to their next `dependents` (natural time flow for calibration),
        defining our `starts` and `ends` of the calibration, as:

        - `starts`: `roots` (``in_degree=0``, no arrows pointing in, no `dependencies`) of the graph, where all arrows leave the node.

        - `ends`: `leaves` (``out_degree=0``, no arrows pointing out, no `dependants`) of the graph, where all arrows enter the node.

        .. code-block:: python

            #                     /--> 2 -->\\
            #                    /     |     \\
            #    (start, root)  0      |      3 --> 4 (end, leave)
            #                    \\     v     /
            #                     \\--> 1 -->/
        """

        self.node_sequence: dict[str, CalibrationNode] = node_sequence
        """Mapping for the nodes of the graph, from strings into the actual initialized nodes (dict)."""

        self.runcard: str = runcard
        """The runcard path, containing the serialized platform where the experiments will be run (str)."""

        self.platform: Platform = build_platform(runcard)
        """The initialized platform, where the experiments will be run (Platform)."""

        self.drift_timeout: float = drift_timeout
        """Duration in seconds, representing an estimate of how long it takes for the calibration parameters to drift.
        During that time the parameters of this node should be considered calibrated.

        Thus a big value will tend to skip recently calibrated nodes, making the calibration process faster, but less accurate,
        and a small value will make the calibration process slower, but more accurate and robust.

        A node will be skipped if the ``drift timeout`` is bigger than the time since its last calibration. Defaults to 7200 (2h).
        """

    def calibrate_all(self, node: CalibrationNode):
        """Calibrates all the nodes sequentially.

        Args:
            node (CalibrationNode): The node where we want to start the `calibration_all()` on. Normally you would want
                this node to be the furthest node in the calibration graph.
        """
        logger.info("WORKFLOW: Calibrating all %s.\n", node.node_id)
        for n in self._dependencies(node):
            self.calibrate_all(n)

        # You can skip it from the `drift_timeout`, but also skip it due to `been_calibrated()`
        # If you want to start the calibration from the start again, just decrease the `drift_timeout` or remove the executed files!
        if not node.been_calibrated:
            if node.previous_timestamp is None or self._is_timeout_expired(node.previous_timestamp, self.drift_timeout):
                self.calibrate(node)
                self._update_parameters(node)

            node.been_calibrated = True
        # After passing this block `node.been_calibrated` will always be True, so it will not be recalibrated again.

    def run_automatic_calibration(self) -> dict[str, dict]:
        """Runs the full automatic calibration procedure and retrieves the final set parameters and achieved fidelities dictionaries.

        This is the primary interface for our calibration procedure and the highest level algorithm, which finds all the end nodes of the graph
        (`leaves`, those without further `dependents`) and runs ``calibrate_all()`` on them.

        Returns:
            dict[str, dict]: Dictionary for the last set parameters and the last achieved fidelities. It contains two dictionaries (dict[tuple, tuple]) in the keys:
                - "set_parameters": Set parameters dictionary, with the key and values being tuples containing, in this order:
                    - key: (``str``: parameter name, ``str``: bus alias, int: qubit).
                    - value: (``float``: parameter value, ``str``: ``node_id`` where computed, ``datetime``: updated time).

                - "fidelities": Fidelities dictionary, with the key and values being tuples containing, in this order:
                    - key: (``str``: parameter name, ``int``: qubit).
                    - value: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        highest_level_nodes = [
            self.node_sequence[node] for node, out_degree in self.calibration_graph.out_degree() if out_degree == 0
        ]

        for node in highest_level_nodes:
            self.calibrate_all(node)

        logger.info(
            "\n#############################################\n"
            "Automatic calibration completed successfully!\n"
            "#############################################\n"
        )
        return self.get_qubit_fidelities_and_parameters_df_tables()

    def get_qubit_fidelities_and_parameters_df_tables(self) -> dict[str, pd.DataFrame]:
        """Generates the 1q, 2q, fidelities and parameters dataframes, with the last calibrations.

        Returns:
            dict[str, pd.DataFrame]: Last calibrations dataframes.
        """
        df_1q, df_2q = self.get_qubits_tables()

        return {
            "1q_table": df_1q,
            "2q_table": df_2q,
            "set_parameters": self.get_last_set_parameters(),
            "fidelities": self.get_last_fidelities(),
        }

    def calibrate(self, node: CalibrationNode) -> None:
        """Runs a node's experiment on its default values of the ``sweep_interval``.

        This method is responsible for calibrating a node to bring it within spec. The calibration process is node-specific
        and depends on the notebook implementation and the calibration parameters.

        .. note:: Find more information about the ``calibrate()`` idea at https://arxiv.org/abs/1803.03226.

        Args:
            node (CalibrationNode): The node where the calibration experiment is run.
        """
        logger.info('WORKFLOW: Calibrating node "%s".\n', node.node_id)
        node.previous_timestamp = node.run_node()
        node._add_string_to_checked_nb_name("calibrated", node.previous_timestamp)
        # add _calibrated tag to the file name, which doesn't have a tag.

    def _update_parameters(self, node: CalibrationNode) -> None:
        """Updates the node parameters value in the platform, after a calibration.

        If the node does not have an associated parameter, or the parameter attribute of the node is None,
        this function does nothing.

        The `node.output_parameters["platform_parameters"]` needs to be a list with the same order as in
        `platform.set_parameter()`:
            - `param_name`: The name of the parameter to be updated.
            - `param_value`: The value of the parameter to be updated.
            - `bus_alias`: The bus alias where the parameter is located.
            - `channel_id`: The channel id where the parameter is located.

        Args:
            node (CalibrationNode): The node which parameters need to be updated in the platform.
        """
        if node.output_parameters is not None and "platform_parameters" in node.output_parameters:
            for param_name, param_value, bus_alias, channel_id in node.output_parameters["platform_parameters"]:
                logger.info(
                    "Platform updated with: %s, %s, (bus: %s, ch: %s).", param_name, param_value, bus_alias, channel_id
                )
                self.platform.set_parameter(
                    parameter=Parameter(param_name), value=param_value, alias=bus_alias, channel_id=channel_id
                )

            save_platform(self.runcard, self.platform)

    def get_last_set_parameters(self) -> pd.DataFrame:
        """Retrieves the last set parameters of the graph.

        Returns:
            dict[tuple, tuple]: Set parameters dictionary, with the key and values being tuples containing, in this order:
                - ``key``: (``str``: parameter name, ``str``: bus alias, ``int``: qubit).
                - ``value``: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        parameters: dict[tuple, tuple] = {}
        for node in self.node_sequence.values():
            if (
                node.output_parameters is not None
                and node.previous_timestamp is not None
                and "platform_parameters" in node.output_parameters
            ):
                for param_name, param_value, bus_alias, channel_id in node.output_parameters["platform_parameters"]:
                    parameters[param_name, bus_alias, channel_id] = (
                        param_value,
                        node.node_id,
                        datetime.fromtimestamp(node.previous_timestamp),
                    )

        df = pd.DataFrame.from_dict(parameters).transpose()
        if len(df.columns) == 3:
            df.columns = ["value", "node_id", "datetime"]
        if df.index.nlevels == 3:
            df.index.names = ["parameter", "bus", "qubit"]
        return df

    def get_last_fidelities(self) -> pd.DataFrame:
        """Retrieves the last updated fidelities of the graph.

        Returns:
            dict[tuple, tuple]: Fidelities dictionary, with the key and values being tuples containing, in this order:
                - ``key``: (``str``: parameter name, ``int``: qubit).
                - ``value``: (``float``: parameter value, ``str``: node_id where computed, ``datetime``: updated time).
        """
        fidelities: dict[tuple, tuple] = {}
        for node in self.node_sequence.values():
            if (
                node.output_parameters is not None
                and node.previous_timestamp is not None
                and "fidelities" in node.output_parameters
            ):
                for qubit, fidelity, value in node.output_parameters["fidelities"]:
                    fidelities[fidelity, qubit] = (
                        value,
                        node.node_id,
                        datetime.fromtimestamp(node.previous_timestamp),
                    )

        df = pd.DataFrame.from_dict(fidelities).transpose()
        if len(df.columns) == 3:
            df.columns = ["fidelity", "node_id", "datetime"]
        if df.index.nlevels == 2:
            df.index.names = ["fidelity", "qubit"]
        return df

    def get_qubits_tables(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Retrieves the last updated fidelities of the graph in the 1 qubit and 2 qubit tables.

        Returns:
            tuple[pd.DataFrame]: Tables where columns are each fidelity or parameter, and rows each qubit.
        """
        q1_df, q2_df = self._create_empty_dataframes()

        for node in self.node_sequence.values():
            qubit = self._get_qubit_from_node(node)
            if len(qubit) == 1:
                q1_df = self._fill_qubits_columns(node, qubit, q1_df)
            else:
                q2_df = self._fill_qubits_columns(node, qubit, q2_df)

        # Reorder fidelities to the front of the dataframe:
        q1_df = self._reorder_fidelities(q1_df)
        q2_df = self._reorder_fidelities(q2_df)

        return q1_df, q2_df

    def _create_empty_dataframes(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Creates the structure of the dataframe for the qubits table.

        Returns:
            pd.DataFrame: Empty df, where columns are each fidelity or parameter, and rows each qubit.
        """
        q1_idx, q1_col, q2_idx, q2_col = [], [], [], []  # type: ignore[var-annotated]

        for node in self.node_sequence.values():
            qubit = self._get_qubit_from_node(node)
            if len(qubit) == 1:
                q1_idx, q1_col = self._get_idx_and_columns(node, qubit, q1_idx, q1_col)
            else:
                q2_idx, q2_col = self._get_idx_and_columns(node, qubit, q2_idx, q2_col)

        df_q1 = pd.DataFrame("-", q1_idx, q1_col)
        df_q1.index.name = "qubit"

        df_q2 = pd.DataFrame("-", q2_idx, q2_col)
        df_q2.index.name = "qubit"

        return df_q1, df_q2

    @staticmethod
    def _get_qubit_from_node(node) -> str:
        """Retrieves the qubit from the node_id.

        Args:
            node (CalibrationNode): The node from which the qubit needs to be retrieved.

        Returns:
            str: The qubit corresponding to the node.
        """
        qubits: list = node.node_id.split("_")
        return "_".join(
            [i for i in qubits if any(char == "q" for char in i) and any(char.isdigit() for char in i)]
        ).replace("q", "-")[1:]

    @staticmethod
    def _get_idx_and_columns(node: CalibrationNode, qubit: str, idx: list, col: list) -> tuple[list, list]:
        """Gets the index and columns for creating a dataframe for 1q or 2q.

        Args:
            node (CalibrationNode): Node for which we are currently creating the dataframe idx and columns.
            qubit (str): qubit for which we are currently creating the dataframe idx and columns, corresponding to the node.
            idx (list): idx to be added.
            col (list): column to be added.

        Returns:
            tuple[list, list]: list of idx and columns for creating the dataframe.
        """
        if qubit not in idx:
            idx.append(qubit)

        if node.output_parameters is not None and node.previous_timestamp is not None:
            if "fidelities" in node.output_parameters:
                for _, fidelity, _ in node.output_parameters["fidelities"]:
                    if fidelity not in col:
                        col.append(fidelity)

            if "platform_parameters" in node.output_parameters:
                for param_name, _, bus_alias, _ in node.output_parameters["platform_parameters"]:
                    bus = CalibrationController._get_bus_name_from_alias(bus_alias)
                    if f"{param_name!s}_{bus}" not in col:
                        col.append(f"{param_name!s}_{bus}")

        return idx, col

    @staticmethod
    def _fill_qubits_columns(node: CalibrationNode, qubit: str, df: pd.DataFrame) -> pd.DataFrame:
        """Fills the qubit tables columns with the fidelities or parameters values.

        Args:
            node (CalibrationNode): Node currently being filled.
            qubit (str): qubit being filled, corresponding to the node.
            df (pd.DataFrame): Dataframe to fill.

        Returns:
            pd.DataFrame: Filled dataframe.
        """
        if node.output_parameters is not None and node.previous_timestamp is not None:
            if "fidelities" in node.output_parameters:
                for _, fidelity, value in node.output_parameters["fidelities"]:
                    df[fidelity][qubit] = value

            if "platform_parameters" in node.output_parameters:
                for param_name, param_value, bus_alias, _ in node.output_parameters["platform_parameters"]:
                    bus = CalibrationController._get_bus_name_from_alias(bus_alias)
                    df[f"{param_name!s}_{bus}"][qubit] = param_value

        return df

    @staticmethod
    def _get_bus_name_from_alias(bus_alias) -> str:
        """Gets the bus name from the bus alias.

        Args:
            bus_alias (str): Bus alias to get the bus name from.

        Returns:
            str: The bus name.
        """
        buses: list = str(bus_alias).split("_")
        return "_".join([x for x in buses if not any(char.isdigit() for char in x)])

    @staticmethod
    def _reorder_fidelities(df: pd.DataFrame) -> pd.DataFrame:
        """It moves the fidelities columns to the front of the dataframe.

        Args:
            df (pd.DataFrame): Dataframe to reorder.

        Returns:
            pd.DataFrame: Reordered dataframe.
        """
        for column in df.columns:
            if "fidelity" in column:
                first_column = df.pop(column)
                df.insert(0, column, first_column)

        return df

    def _dependencies(self, node: CalibrationNode) -> list:
        """Finds the dependencies of a node.

        If in our graph we have `A -> B`, then node B depends on node A. Thus calling this method on B would return A.

        Args:
            node (CalibrationNode): The nodes for which the dependencies need to be retrieved.

        Returns:
            list: The nodes that the argument node depends on.
        """
        return [self.node_sequence[node_name] for node_name in self.calibration_graph.predecessors(node.node_id)]

    @staticmethod
    def _is_timeout_expired(timestamp: float, timeout: float) -> bool:
        """Checks if the time passed since the ``timestamp`` is greater than the ``timeout`` duration.

        Args:
            timestamp (float): Timestamp from which the time should be checked, described in UNIX timestamp format.
            timeout (float): The timeout duration in seconds.

        Returns:
            bool: True if the ``timeout`` has expired, False otherwise.
        """
        # Calculate the time that should have passed (timestamp + timeout duration), convert them to datetime objects:
        timestamp_dt = datetime.fromtimestamp(timestamp)
        timeout_duration = timedelta(seconds=timeout)
        timeout_time = timestamp_dt + timeout_duration

        # Check if the current time is greater than the timeout time
        current_time = datetime.now()
        return current_time > timeout_time



import numpy as np
import pytest

from qililab.waveforms import Gaussian


@pytest.fixture(name="gaussian")
def fixture_gaussian():
    return Gaussian(amplitude=1, duration=10, num_sigmas=2.5)


class TestGaussian:
    def test_init(self, gaussian):
        """Test __init__ method"""
        assert gaussian.amplitude == 1
        assert gaussian.duration == 10
        assert gaussian.num_sigmas == 2.5

    def test_envelope(self, gaussian):
        """Test envelope method"""

        # calculate envelope
        sigma = gaussian.duration / gaussian.num_sigmas
        mu = gaussian.duration / 2
        x = np.arange(5) * 2
        envelope = np.exp(-0.5 * (x - mu) ** 2 / sigma**2)
        norm = np.amax(np.real(envelope))
        envelope = envelope - envelope[0]
        corr_norm = np.amax(np.real(envelope))
        envelope = envelope * norm / corr_norm

        assert np.allclose(gaussian.envelope(resolution=2), envelope)

    def test_envelope_method_with_zero_amplitude(self, gaussian):
        """Test envelope method when amplitude is zero."""
        gaussian.amplitude = 0
        assert np.allclose(gaussian.envelope(), np.zeros(10))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the QbloxQCMRF class."""

from dataclasses import dataclass, field
from typing import ClassVar, Optional

from qblox_instruments.qcodes_drivers.module import Module as QcmQrm

from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue

from .qblox_qrm import QbloxQRM


@InstrumentFactory.register
class QbloxQRMRF(QbloxQRM):
    """Qblox QRM-RF driver."""

    name = InstrumentName.QRMRF
    device: QcmQrm

    @dataclass
    class QbloxQRMRFSettings(QbloxQRM.QbloxQRMSettings):
        """Contains the settings of a specific Qblox QRM-RF module."""

        out0_in0_lo_freq: float
        out0_in0_lo_en: bool
        out0_att: int  # must be a multiple of 2!
        in0_att: int  # must be a multiple of 2!
        out0_offset_path0: float
        out0_offset_path1: float
        out_offsets: list[float] = field(init=False, default_factory=list)
        out0_in0_lo_freq_cal_type_default: Optional[str] = "off"

    # TODO: We should separate instrument settings and instrument parameters, such that the user can quickly get
    # al the settable parameters of an instrument.
    parameters: ClassVar[set[Parameter]] = {
        Parameter.OUT0_IN0_LO_FREQ,
        Parameter.OUT0_IN0_LO_EN,
        Parameter.OUT0_ATT,
        Parameter.IN0_ATT,
        Parameter.OUT0_OFFSET_PATH0,
        Parameter.OUT0_OFFSET_PATH1,
        Parameter.OUT0_IN0_LO_FREQ_CAL_TYPE_DEFAULT,
    }

    settings: QbloxQRMRFSettings

    @check_device_initialized
    def initial_setup(self):
        """Initial setup"""
        super().initial_setup()
        for parameter in self.parameters:
            self.set_parameter(parameter, getattr(self.settings, parameter.value))

    def _map_connections(self):
        """Disable all connections and map sequencer paths with output/input channels."""
        # Disable all connections
        self.device.disconnect_outputs()
        self.device.disconnect_inputs()

        for sequencer_dataclass in self.awg_sequencers:
            sequencer = self.device.sequencers[sequencer_dataclass.identifier]
            sequencer.connect_out0("IQ")
            sequencer.connect_acq("in0")

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set a parameter of the Qblox QCM-RF module.
        Args:
            parameter (Parameter): Parameter name.
            value (float | str | bool): Value to set.
            channel_id (int, optional): ID of the sequencer. Defaults to None.
        """
        if parameter == Parameter.LO_FREQUENCY:
            parameter = Parameter.OUT0_IN0_LO_FREQ

        if parameter == Parameter.OUT0_ATT:
            max_att = self.device._get_max_out_att_0()
            if value > max_att:
                raise Exception(
                    f"`{Parameter.OUT0_ATT}` for this module cannot be higher than {max_att}dB.\n"
                    "Please specify an attenuation level, multiple of 2, below this value."
                )

        if parameter in self.parameters:
            setattr(self.settings, parameter.value, value)

            if self.is_device_active():
                self.device.set(parameter.value, value)
            return
        super().set_parameter(parameter, value, channel_id)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get a parameter of the Qblox QRM-RF module.
        Args:
            parameter (Parameter): Parameter name.
            value (float | str | bool): Value to set.
            channel_id (int, optional): ID of the sequencer. Defaults to None.
        """
        if parameter == Parameter.LO_FREQUENCY:
            parameter = Parameter.OUT0_IN0_LO_FREQ

        if parameter in self.parameters:
            return getattr(self.settings, parameter.value)
        return super().get_parameter(parameter, channel_id)

    def to_dict(self):
        """Return a dict representation of an `QRM-RF` instrument."""
        dictionary = super().to_dict()
        dictionary.pop("out_offsets")
        return dictionary

    def calibrate_mixers(self, cal_type: str, channel_id: ChannelID | None = None):
        if cal_type == "lo":
            self.device._run_mixer_lo_calib(channel_id)
        elif cal_type == "lo_and_sidebands":
            self.device._run_mixer_lo_calib(channel_id)
            for sequencer_dataclass in self.awg_sequencers:
                sequencer = self.device.sequencers[sequencer_dataclass.identifier]
                sequencer.sideband_cal()
        else:
            raise Exception(
                f"`{cal_type}` for this module must be one of the following values: " "'lo' or 'lo_and_sidebands'."
            )

        return



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This module contains a function to display all the details of the Qililab installation.
"""

import platform
import sys

import pyvisa_py
import qblox_instruments
import qcodes
import qcodes_contrib_drivers
import qibo
import qpysequence
from qm.version import __version__ as qm_version


def about():
    """
    Prints the information for Qililab installation.
    """
    print(f"Platform info:             {platform.platform(aliased=True)}")
    print(f"Python version:            {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}")
    print(f"PyVISA version:            {pyvisa_py.__version__}")
    print(f"QCodes version:            {qcodes.__version__}")
    print(f"QCodes Contrib version:    {qcodes_contrib_drivers.__version__}")
    print(f"Qblox Instrument version:  {qblox_instruments.__version__}")
    print(f"Qpysequence version:       {qpysequence.__version__}")
    print(f"Quantum Machines version:  {qm_version}")
    print(f"Qibo version:              {qibo.__version__}")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .acquire import Acquire, AcquireWithCalibratedWeights
from .execute_qprogram import ExecuteQProgram
from .get_parameter import GetParameter
from .measure import (
    Measure,
    MeasureWithCalibratedWaveform,
    MeasureWithCalibratedWaveformWeights,
    MeasureWithCalibratedWeights,
)
from .operation import Operation
from .play import Play, PlayWithCalibratedWaveform
from .reset_phase import ResetPhase
from .set_crosstalk import SetCrosstalk
from .set_frequency import SetFrequency
from .set_gain import SetGain
from .set_markers import SetMarkers
from .set_offset import SetOffset
from .set_parameter import SetParameter
from .set_phase import SetPhase
from .sync import Sync
from .wait import Wait

__all__ = [
    "Acquire",
    "AcquireWithCalibratedWeights",
    "ExecuteQProgram",
    "GetParameter",
    "Measure",
    "MeasureWithCalibratedWaveform",
    "MeasureWithCalibratedWaveformWeights",
    "MeasureWithCalibratedWeights",
    "Operation",
    "Play",
    "PlayWithCalibratedWaveform",
    "ResetPhase",
    "SetCrosstalk",
    "SetFrequency",
    "SetGain",
    "SetMarkers",
    "SetOffset",
    "SetParameter",
    "SetPhase",
    "Sync",
    "Wait",
]



import numpy as np
import pytest

from qililab import Arbitrary, Calibration, Domain, IQPair, QProgram, QuantumMachinesCompiler, Square
from qililab.qprogram.blocks import ForLoop, Loop


@pytest.fixture(name="calibration")
def fixture_calibration() -> Calibration:
    calibration = Calibration()
    calibration.add_waveform(bus="drive_q0", name="Xpi", waveform=Square(1.0, 100))
    calibration.add_waveform(bus="drive_q1", name="Xpi", waveform=Square(1.0, 150))
    calibration.add_waveform(bus="drive_q2", name="Xpi", waveform=Square(1.0, 200))

    return calibration


@pytest.fixture(name="play_operation")
def fixture_play_operation() -> QProgram:
    ones_zeros_pair = IQPair(I=Square(amplitude=1.0, duration=100), Q=Square(amplitude=0.0, duration=100))
    qp = QProgram()
    qp.play(bus="drive", waveform=ones_zeros_pair)

    return qp


@pytest.fixture(name="measurement_blocked_operation")
def fixture_measurement_blocked_operation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    with qp.block():
        qp.play(bus="drive", waveform=drag_wf)
        qp.measure(bus="readout", waveform=readout_pair, weights=weights_pair)

    return qp


@pytest.fixture(name="play_operations_share_waveforms")
def fixture_play_operations_share_waveforms() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    qp = QProgram()
    qp.play(bus="drive", waveform=drag_wf)
    qp.play(bus="drive", waveform=IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5))

    return qp


@pytest.fixture(name="set_gain_and_play_operation")
def fixture_set_gain_and_play_operation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    qp = QProgram()
    qp.set_gain(bus="drive", gain=0.5)
    qp.play(bus="drive", waveform=drag_wf)

    return qp


@pytest.fixture(name="play_named_operation")
def fixture_play_named_operation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    qp = QProgram()
    qp.play(bus="drive", waveform="Xpi")
    qp.play(bus="drive", waveform=drag_wf)

    return qp


@pytest.fixture(name="set_frequency_operation")
def fixture_set_frequency_operation() -> QProgram:
    qp = QProgram()
    qp.set_frequency(bus="drive", frequency=100e6)

    return qp


@pytest.fixture(name="set_offset_operation")
def fixture_set_offset_operation() -> QProgram:
    qp = QProgram()
    qp.set_offset(bus="drive", offset_path0=0.1, offset_path1=0.2)

    return qp


@pytest.fixture(name="set_dc_offset_operation")
def fixture_set_dc_offset_operation() -> QProgram:
    qp = QProgram()
    qp.set_offset(bus="drive", offset_path0=0.1)

    return qp


@pytest.fixture(name="set_phase_operation")
def fixture_set_phase_operation() -> QProgram:
    qp = QProgram()
    qp.set_phase(bus="drive", phase=np.pi / 2)

    return qp


@pytest.fixture(name="reset_phase_operation")
def fixture_reset_phase_operation() -> QProgram:
    qp = QProgram()
    qp.reset_phase(bus="drive")

    return qp


@pytest.fixture(name="wait_operation")
def fixture_wait_operation() -> QProgram:
    qp = QProgram()
    qp.wait(bus="drive", duration=100)

    return qp


@pytest.fixture(name="sync_operation")
def fixture_sync_operation() -> QProgram:
    qp = QProgram()
    qp.sync(buses=["drive", "readout"])

    return qp


@pytest.fixture(name="sync_operation_no_parameters")
def fixture_sync_operation_no_parameters() -> QProgram:
    qp = QProgram()
    qp.wait(bus="drive", duration=100)
    qp.wait(bus="readout", duration=200)
    qp.sync()

    return qp


@pytest.fixture(name="measure_operation")
def fixture_measure_operation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    qp.measure(bus="readout", waveform=drag_wf, weights=weights)

    return qp


@pytest.fixture(name="measure_operation_save_raw_adc")
def fixture_measure_operation_save_raw_adc() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    qp.quantum_machines.measure(bus="readout", waveform=drag_wf, weights=weights, save_adc=True)

    return qp


@pytest.fixture(name="measure_operation_no_demodulation")
def fixture_measure_operation_no_demodulation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    qp.quantum_machines.measure(bus="readout", waveform=drag_wf, weights=weights, demodulation=False)

    return qp


@pytest.fixture(name="measure_operation_with_same_pulse")
def fixture_measure_operation_with_same_pulse() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    weights2 = IQPair(I=Square(0.5, duration=200), Q=Square(0.5, duration=200))

    qp = QProgram()
    qp.quantum_machines.measure(bus="readout", waveform=drag_wf, weights=weights, demodulation=False)
    qp.quantum_machines.measure(bus="readout", waveform=drag_wf, weights=weights2, demodulation=False)

    return qp


@pytest.fixture(name="measure_operation_with_average")
def fixture_measure_operation_with_average() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    with qp.average(shots=1000):
        qp.measure(bus="readout", waveform=drag_wf, weights=weights)

    return qp


@pytest.fixture(name="measure_operation_with_inner_loop_average")
def fixture_measure_operation_with_inner_loop_averagee() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.for_loop(variable=gain, start=0, stop=1.0, step=0.1):
        qp.set_gain(bus="readout", gain=gain)
        with qp.average(shots=1000):
            qp.measure(bus="readout", waveform=drag_wf, weights=weights)

    return qp


@pytest.fixture(name="measure_operation_in_for_loop")
def fixture_measure_operation_in_for_loop() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.for_loop(variable=gain, start=0, stop=1.0, step=0.1):
        qp.set_gain(bus="readout", gain=gain)
        qp.measure(bus="readout", waveform=drag_wf, weights=weights)

    return qp


@pytest.fixture(name="measure_operation_in_loop")
def fixture_measure_operation_in_loop() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.loop(variable=gain, values=np.arange(start=0, stop=1.05, step=0.1)):
        qp.set_gain(bus="readout", gain=gain)
        qp.measure(bus="readout", waveform=drag_wf, weights=weights)

    return qp


@pytest.fixture(name="measure_operation_in_parallel")
def fixture_measure_operation_in_parallel() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    weights = IQPair(I=Square(1.0, duration=200), Q=Square(1.0, duration=200))
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    with qp.parallel(
        loops=[
            Loop(variable=gain, values=np.arange(start=0, stop=1.05, step=0.1)),
            Loop(variable=frequency, values=np.arange(start=100, stop=205, step=10)),
        ]
    ):
        qp.set_frequency(bus="readout", frequency=frequency)
        qp.set_gain(bus="readout", gain=gain)
        qp.measure(bus="readout", waveform=drag_wf, weights=weights)

    return qp


@pytest.fixture(name="for_loop")
def fixture_for_loop() -> QProgram:
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    phase = qp.variable(label="phase", domain=Domain.Phase)
    time = qp.variable(label="time", domain=Domain.Time)
    scalar = qp.variable(label="int_scalar", domain=Domain.Scalar, type=int)

    with qp.for_loop(variable=gain, start=0, stop=1.0, step=0.1):
        qp.set_gain(bus="drive", gain=gain)

    with qp.for_loop(variable=frequency, start=100, stop=200, step=10):
        qp.set_frequency(bus="drive", frequency=frequency)

    with qp.for_loop(variable=phase, start=0, stop=np.pi / 2, step=np.pi / 18):
        qp.set_phase(bus="drive", phase=phase)

    with qp.for_loop(variable=time, start=100, stop=200, step=10):
        qp.wait(bus="drive", duration=time)

    with qp.for_loop(variable=scalar, start=0, stop=10, step=1):
        qp.wait(bus="drive", duration=100)

    return qp


@pytest.fixture(name="for_loop_with_negative_step")
def fixture_for_loop_with_negative_step() -> QProgram:
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    phase = qp.variable(label="phase", domain=Domain.Phase)
    time = qp.variable(label="time", domain=Domain.Time)
    scalar = qp.variable(label="int_scalar", domain=Domain.Scalar, type=int)

    with qp.for_loop(variable=gain, start=1.0, stop=0.0, step=-0.1):
        qp.set_gain(bus="drive", gain=gain)

    with qp.for_loop(variable=frequency, start=200, stop=100, step=-10):
        qp.set_frequency(bus="drive", frequency=frequency)

    with qp.for_loop(variable=phase, start=np.pi / 2, stop=0, step=-np.pi / 18):
        qp.set_phase(bus="drive", phase=phase)

    with qp.for_loop(variable=time, start=200, stop=100, step=-10):
        qp.wait(bus="drive", duration=time)

    with qp.for_loop(variable=scalar, start=10, stop=0, step=-1):
        qp.wait(bus="drive", duration=100)

    return qp


@pytest.fixture(name="loop")
def fixture_loop() -> QProgram:
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    phase = qp.variable(label="phase", domain=Domain.Phase)
    time = qp.variable(label="time", domain=Domain.Time)
    scalar = qp.variable(label="int_scalar", domain=Domain.Scalar, type=int)

    with qp.loop(variable=gain, values=np.arange(start=0, stop=1.05, step=0.1)):
        qp.set_gain(bus="drive", gain=gain)

    with qp.loop(variable=frequency, values=np.arange(start=100, stop=205, step=10)):
        qp.set_frequency(bus="drive", frequency=frequency)

    with qp.loop(variable=phase, values=np.arange(start=0, stop=95, step=10)):
        qp.set_phase(bus="drive", phase=phase)

    with qp.loop(variable=time, values=np.arange(start=100, stop=205, step=10)):
        qp.wait(bus="drive", duration=time)

    with qp.loop(variable=scalar, values=np.arange(start=0, stop=10, step=1)):
        qp.wait(bus="drive", duration=100)

    return qp


@pytest.fixture(name="parallel")
def fixture_parallel() -> QProgram:
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    phase = qp.variable(label="phase", domain=Domain.Phase)
    time = qp.variable(label="time", domain=Domain.Time)
    scalar = qp.variable(label="int_scalar", domain=Domain.Scalar, type=int)

    with qp.parallel(
        loops=[
            ForLoop(variable=gain, start=0, stop=1.0, step=0.1),
            Loop(variable=frequency, values=np.arange(start=100, stop=205, step=10)),
            ForLoop(variable=phase, start=0, stop=90, step=10),
            Loop(variable=time, values=np.arange(start=100, stop=205, step=10)),
            ForLoop(variable=scalar, start=0, stop=10, step=1),
        ]
    ):
        qp.set_gain(bus="drive", gain=gain)
        qp.set_frequency(bus="drive", frequency=frequency)
        qp.set_phase(bus="drive", phase=phase)
        qp.wait(bus="drive", duration=time)

    return qp


@pytest.fixture(name="infinite_loop")
def fixture_infinite_loop() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    qp = QProgram()
    with qp.infinite_loop():
        qp.play(bus="drive", waveform=drag_wf)

    return qp


class TestQuantumMachinesCompiler:
    def test_play_operation(self, play_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, _ = compiler.compile(play_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        play = statements[0].play
        assert play.qe.name == "drive"
        assert play.named_pulse.name in configuration["pulses"]

    def test_block_handlers(self, measurement_blocked_operation: QProgram):
        drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
        readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
        weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
        qp_no_block = QProgram()
        qp_no_block.play(bus="drive", waveform=drag_wf)
        qp_no_block.measure(bus="readout", waveform=readout_pair, weights=weights_pair)

        compiler = QuantumMachinesCompiler()
        qua_program, configuration, _ = compiler.compile(qprogram=measurement_blocked_operation)

        qua_program_no_block, _, _ = compiler.compile(qprogram=qp_no_block)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 4

        play = statements[0].play
        assert play.qe.name == "drive"
        assert play.named_pulse.name in configuration["pulses"]

        measure = statements[1].measure
        assert measure.qe.name == "readout"
        assert measure.pulse.name in configuration["pulses"]

        assert qua_program._program == qua_program_no_block._program

    def test_play_operations_share_waveforms(self, play_operations_share_waveforms: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, _ = compiler.compile(play_operations_share_waveforms)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 2

        play1 = statements[0].play
        assert play1.qe.name == "drive"
        assert play1.named_pulse.name in configuration["pulses"]

        play2 = statements[0].play
        assert play2.qe.name == "drive"
        assert play2.named_pulse.name in configuration["pulses"]

        assert play1.named_pulse.name == play2.named_pulse.name

    def test_set_gain_and_play_operation(self, set_gain_and_play_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, _ = compiler.compile(set_gain_and_play_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        play = statements[0].play
        assert play.qe.name == "drive"
        assert play.named_pulse.name in configuration["pulses"]
        assert float(play.amp.v0.literal.value) == 1

    def test_play_named_operation_and_bus_mapping(self, play_named_operation: QProgram, calibration: Calibration):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(
            play_named_operation, bus_mapping={"drive": "drive_q0"}, calibration=calibration
        )

        statements = qua_program._program.script.body.statements
        assert len(statements) == 2

        play1 = statements[0].play
        assert play1.qe.name == "drive_q0"

        play2 = statements[1].play
        assert play2.qe.name == "drive_q0"

    def test_play_named_operation_raises_error_if_operations_not_in_calibration(self, play_named_operation: QProgram):
        calibration = Calibration()
        compiler = QuantumMachinesCompiler()
        with pytest.raises(RuntimeError):
            _, _, _ = compiler.compile(play_named_operation, bus_mapping={"drive": "drive_q0"}, calibration=calibration)

    def test_set_frequency_operation(self, set_frequency_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(set_frequency_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        update_frequency = statements[0].update_frequency
        assert update_frequency.qe.name == "drive"
        assert update_frequency.keep_phase is False
        assert float(update_frequency.value.literal.value) == 100e6

    def test_set_offset_operation(self, set_offset_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(set_offset_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 2

        set_dc_offset = statements[0].set_dc_offset
        assert set_dc_offset.qe.name == "drive"

    def test_set_dc_offset_operation(self, set_dc_offset_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(set_dc_offset_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        set_dc_offset = statements[0].set_dc_offset
        assert set_dc_offset.qe.name == "drive"

    def test_set_phase_operation(self, set_phase_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(set_phase_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        rotation = statements[0].z_rotation
        assert rotation.qe.name == "drive"
        assert float(rotation.value.literal.value) == 90 / 360.0

    def test_reset_phase_operation(self, reset_phase_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(reset_phase_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        reset_frame = statements[0].reset_frame
        assert reset_frame.qe.name == "drive"

    def test_wait_operation(self, wait_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(wait_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        wait = statements[0].wait
        assert len(wait.qe) == 1
        assert wait.qe[0].name == "drive"
        assert int(wait.time.literal.value) == 100 / 4

    def test_sync_operation(self, sync_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(sync_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        align = statements[0].align
        assert len(align.qe) == 2
        assert align.qe[0].name == "drive"
        assert align.qe[1].name == "readout"

    def test_sync_operation_no_parameters(self, sync_operation_no_parameters: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(sync_operation_no_parameters)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 3

        align = statements[2].align
        assert len(align.qe) == 0

    def test_measure_operation(self, measure_operation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, measurements = compiler.compile(measure_operation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 3

        measure = statements[0].measure
        assert measure.qe.name == "readout"
        assert measure.pulse.name in configuration["pulses"]

        assert len(measure.measure_processes) == 2
        assert measure.measure_processes[0].analog.dual_demod_integration.element_output1 == "out1"
        assert measure.measure_processes[0].analog.dual_demod_integration.element_output2 == "out2"
        assert measure.measure_processes[1].analog.dual_demod_integration.element_output1 == "out1"
        assert measure.measure_processes[1].analog.dual_demod_integration.element_output2 == "out2"

        measurement_pulse = configuration["pulses"][measure.pulse.name]
        assert len(measurement_pulse["integration_weights"]) == 4

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_save_raw_adc(self, measure_operation_save_raw_adc: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, measurements = compiler.compile(measure_operation_save_raw_adc)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 3

        measure = statements[0].measure
        assert measure.qe.name == "readout"
        assert measure.pulse.name in configuration["pulses"]

        assert len(measure.measure_processes) == 2
        assert measure.measure_processes[0].analog.dual_demod_integration.element_output1 == "out1"
        assert measure.measure_processes[0].analog.dual_demod_integration.element_output2 == "out2"
        assert measure.measure_processes[1].analog.dual_demod_integration.element_output1 == "out1"
        assert measure.measure_processes[1].analog.dual_demod_integration.element_output2 == "out2"

        measurement_pulse = configuration["pulses"][measure.pulse.name]
        assert len(measurement_pulse["integration_weights"]) == 4

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 4
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles
        assert "adc1_0" in measurements[0].result_handles
        assert "adc2_0" in measurements[0].result_handles

    def test_measure_operation_no_demodulation(self, measure_operation_no_demodulation: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, measurements = compiler.compile(measure_operation_no_demodulation)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 3

        measure = statements[0].measure
        assert measure.qe.name == "readout"
        assert measure.pulse.name in configuration["pulses"]

        assert len(measure.measure_processes) == 2
        assert measure.measure_processes[0].analog.dual_bare_integration.element_output1 == "out1"
        assert measure.measure_processes[1].analog.dual_bare_integration.element_output1 == "out1"
        assert measure.measure_processes[1].analog.dual_bare_integration.element_output2 == "out2"
        assert measure.measure_processes[0].analog.dual_bare_integration.element_output2 == "out2"

        measurement_pulse = configuration["pulses"][measure.pulse.name]
        assert len(measurement_pulse["integration_weights"]) == 4

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_with_same_pulse_updates_it_correctly(self, measure_operation_with_same_pulse: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, configuration, _ = compiler.compile(measure_operation_with_same_pulse)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 6

        measure_0 = statements[0].measure
        assert measure_0.qe.name == "readout"
        assert measure_0.pulse.name in configuration["pulses"]

        measure_1 = statements[3].measure
        assert measure_1.qe.name == "readout"
        assert measure_1.pulse.name in configuration["pulses"]

        assert measure_0.pulse.name == measure_1.pulse.name

        assert len(configuration["pulses"][measure_0.pulse.name]["integration_weights"]) == 8

    def test_measure_operation_with_average(self, measure_operation_with_average: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, measurements = compiler.compile(measure_operation_with_average)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_with_inner_loop_average(self, measure_operation_with_inner_loop_average: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, measurements = compiler.compile(measure_operation_with_inner_loop_average)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_in_for_loop(self, measure_operation_in_for_loop: QProgram):
        compiler = QuantumMachinesCompiler()
        _, _, measurements = compiler.compile(measure_operation_in_for_loop)

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_in_loop(self, measure_operation_in_loop: QProgram):
        compiler = QuantumMachinesCompiler()
        _, _, measurements = compiler.compile(measure_operation_in_loop)

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_in_parallel(self, measure_operation_in_parallel: QProgram):
        compiler = QuantumMachinesCompiler()
        _, _, measurements = compiler.compile(measure_operation_in_parallel)

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_measure_operation_with_threshold(self, measure_operation: QProgram):
        """Test measurement result contains the provided thresholds."""
        compiler = QuantumMachinesCompiler()
        thresholds = {"readout": 1.0}
        _, _, measurements = compiler.compile(measure_operation, thresholds=thresholds)

        assert len(measurements) == 1
        assert measurements[0].threshold == 1.0

    def test_measure_operation_with_threshold_rotations(self, measure_operation: QProgram):
        """Test compilation of measurement applying the rotations provided in the `threshold_rotations` map"""
        compiler = QuantumMachinesCompiler()
        rotation_angle = np.pi
        threshold_rotations = {"readout": rotation_angle}
        qua_program, configuration, measurements = compiler.compile(
            measure_operation, threshold_rotations=threshold_rotations
        )

        statements = qua_program._program.script.body.statements
        assert len(statements) == 3

        measure = statements[0].measure
        assert measure.qe.name == "readout"
        assert measure.pulse.name in configuration["pulses"]

        assert len(measure.measure_processes) == 2
        assert measure.measure_processes[0].analog.dual_demod_integration.element_output1 == "out1"
        assert measure.measure_processes[0].analog.dual_demod_integration.element_output2 == "out2"
        assert measure.measure_processes[1].analog.dual_demod_integration.element_output1 == "out1"
        assert measure.measure_processes[1].analog.dual_demod_integration.element_output2 == "out2"

        measurement_pulse = configuration["pulses"][measure.pulse.name]
        assert len(measurement_pulse["integration_weights"]) == 4

        A, B, C, D = configuration["integration_weights"].values()
        np.testing.assert_allclose(A["cosine"], [(np.cos(rotation_angle), 200)], atol=1e-15)
        np.testing.assert_allclose(A["sine"], [(np.sin(rotation_angle), 200)], atol=1e-15)

        np.testing.assert_allclose(B["cosine"], [(-np.sin(rotation_angle), 200)], atol=1e-15)
        np.testing.assert_allclose(B["sine"], [(np.cos(rotation_angle), 200)], atol=1e-15)

        np.testing.assert_allclose(C["cosine"], [(np.sin(rotation_angle), 200)], atol=1e-15)
        np.testing.assert_allclose(C["sine"], [(-np.cos(rotation_angle), 200)], atol=1e-15)

        np.testing.assert_allclose(D["cosine"], [(np.cos(rotation_angle), 200)], atol=1e-15)
        np.testing.assert_allclose(D["sine"], [(np.sin(rotation_angle), 200)], atol=1e-15)

        assert len(measurements) == 1
        assert len(measurements[0].result_handles) == 2
        assert "I_0" in measurements[0].result_handles
        assert "Q_0" in measurements[0].result_handles

    def test_for_loop(self, for_loop: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(for_loop)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 5

        # Voltage
        assert float(statements[0].for_.init.statements[0].assign.expression.literal.value) == 0
        assert float(statements[0].for_.condition.binary_operation.right.literal.value) == 1.05
        assert (
            float(statements[0].for_.update.statements[0].assign.expression.binary_operation.right.literal.value) == 0.1
        )

        # Frequency
        assert float(statements[1].for_.init.statements[0].assign.expression.literal.value) == 100
        assert float(statements[1].for_.condition.binary_operation.right.literal.value) == 205
        assert (
            float(statements[1].for_.update.statements[0].assign.expression.binary_operation.right.literal.value) == 10
        )

        # Phase
        assert float(statements[2].for_.init.statements[0].assign.expression.literal.value) == 0 / 360.0
        assert float(statements[2].for_.condition.binary_operation.right.literal.value) == 95 / 360.0
        assert (
            float(statements[2].for_.update.statements[0].assign.expression.binary_operation.right.literal.value)
            == 10 / 360.0
        )

        # Time
        assert float(statements[3].for_.init.statements[0].assign.expression.literal.value) == 100
        assert float(statements[3].for_.condition.binary_operation.right.literal.value) == 200
        assert (
            float(statements[3].for_.update.statements[0].assign.expression.binary_operation.right.literal.value) == 10
        )

    def test_for_loop_with_negative_step(self, for_loop_with_negative_step: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(for_loop_with_negative_step)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 5

        # Voltage
        assert float(statements[0].for_.init.statements[0].assign.expression.literal.value) == 1.0
        assert float(statements[0].for_.condition.binary_operation.right.literal.value) == -0.05
        assert (
            float(statements[0].for_.update.statements[0].assign.expression.binary_operation.right.literal.value)
            == -0.1
        )

        # Frequency
        assert float(statements[1].for_.init.statements[0].assign.expression.literal.value) == 200
        assert float(statements[1].for_.condition.binary_operation.right.literal.value) == 95
        assert (
            float(statements[1].for_.update.statements[0].assign.expression.binary_operation.right.literal.value) == -10
        )

        # Phase
        assert float(statements[2].for_.init.statements[0].assign.expression.literal.value) == 90 / 360.0
        assert float(statements[2].for_.condition.binary_operation.right.literal.value) == -5 / 360.0
        assert (
            float(statements[2].for_.update.statements[0].assign.expression.binary_operation.right.literal.value)
            == -10 / 360.0
        )

        # Time
        assert float(statements[3].for_.init.statements[0].assign.expression.literal.value) == 200
        assert float(statements[3].for_.condition.binary_operation.right.literal.value) == 100
        assert (
            float(statements[3].for_.update.statements[0].assign.expression.binary_operation.right.literal.value) == -10
        )

    def test_loop(self, loop: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(loop)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 5
        assert len(statements[0].for_each.iterator) == 1
        assert len(statements[1].for_each.iterator) == 1
        assert len(statements[2].for_each.iterator) == 1
        assert len(statements[3].for_each.iterator) == 1
        assert len(statements[4].for_each.iterator) == 1

    def test_parallel(self, parallel: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(parallel)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1
        assert len(statements[0].for_each.iterator) == 5

    def test_infinite_loop(self, infinite_loop: QProgram):
        compiler = QuantumMachinesCompiler()
        qua_program, _, _ = compiler.compile(infinite_loop)

        statements = qua_program._program.script.body.statements
        assert len(statements) == 1
        assert bool(statements[0].for_.condition.literal.value) is True

    def test_hash_arbitrary_waveforms(self):
        compiler = QuantumMachinesCompiler()
        waveform0 = Arbitrary(np.ones(3000))
        waveform1 = Arbitrary(np.ones(3004))
        assert compiler._QuantumMachinesCompiler__hash_waveform(
            waveform0
        ) != compiler._QuantumMachinesCompiler__hash_waveform(waveform1)

    @pytest.mark.parametrize(
        "start,stop,step,expected_result",
        [(0, 10, 1, 11), (10, 0, -1, 11), (1, 2.05, 0.1, 11)],
    )
    def test_calculate_iterations(self, start, stop, step, expected_result):
        result = QuantumMachinesCompiler._calculate_iterations(start, stop, step)
        assert result == expected_result

    def test_calculate_iterations_with_zero_step_throws_error(self):
        with pytest.raises(ValueError, match="Step value cannot be zero"):
            QuantumMachinesCompiler._calculate_iterations(100, 200, 0)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Enum classes"""

from enum import Enum

from qililab.yaml import yaml


class Instrument(str, Enum):
    """Instrument.

    Args:
        enum (str): Available types of instruments:
        * platform
        * awg
        * signal_generator
        * attenuator
        * voltage_source
        * current_source
    """

    PLATFORM = "platform"
    AWG = "awg"
    SIGNAL_GENERATOR = "signal_generator"
    ATTENUATOR = "attenuator"
    VOLTAGE_SOURCE = "voltage_source"
    CURRENT_SOURCE = "current_source"


class ReferenceClock(str, Enum):
    """Qblox reference clock.

    Args:
        enum (str): Available types of reference clock:
        * Internal
        * External
    """

    INTERNAL = "internal"
    EXTERNAL = "external"


class AcquireTriggerMode(str, Enum):
    """Qblox acquire trigger mode.

    Args:
        enum (str): Available types of trigger modes:
        * sequencer
        * level
    """

    SEQUENCER = "sequencer"
    LEVEL = "level"


class IntegrationMode(str, Enum):
    """Qblox integration mode.

    Args:
        enum (str): Available types of integration modes:
        * ssb
    """

    SSB = "ssb"


class GateName(str, Enum):
    """Gate names.

    Args:
        enum (str): Available types of gate names:
        * I
        * X
        * Y
        * M
        * RX
        * RY
        * XY
        * Drag
        * CZ
        * Park
    """

    I = "I"
    X = "X"
    RX = "RX"
    Y = "Y"
    RY = "RY"
    XY = "XY"
    M = "M"
    Drag = "Drag"
    CZ = "CZ"
    Park = "Park"


class AcquisitionName(str, Enum):
    """Acquisition names.

    Args:
        enum (str): Available types of acquisition names:
        * single
    """

    SINGLE = "single"
    LARGE = "large"


class PulseDistortionName(str, Enum):
    """Pulse distortion options.

    Args:
        Enum (str): Available types of PulseDistortion options:
        * gaussian
    """

    BIAS_TEE_CORRECTION = "bias_tee"
    EXPONENTIAL_CORRECTION = "exponential"
    LFILTER = "lfilter"


class PulseShapeName(str, Enum):
    """Pulse shape options.

    Args:
        Enum (str): Available types of PulseShape options:
        * gaussian
    """

    GAUSSIAN = "gaussian"
    DRAG = "drag"
    RECTANGULAR = "rectangular"
    SNZ = "snz"
    COSINE = "cosine"
    FLATTOP = "flat_top"
    TWOSTEP = "two_step"


class InstrumentName(str, Enum):
    """Instrument names.

    Args:
        enum (str): Available instrument element names:
        * QCM -> Exactly as Qblox InstrumentType
        * QRM -> Exactly as Qblox InstrumentType
        * rohde_schwarz
        * mini_circuits
        * keithley_2600
        * qblox_D5a
        * qblox_S4g
        * keysight_e5080b
        * agilent_e5071B
        * yokogawa_gs200
        * OPX -> Exactly as Quantum Machines InstrumentType
    """

    QBLOX_QCM = "QCM"
    QBLOX_QRM = "QRM"
    QRMRF = "QRM-RF"
    ROHDE_SCHWARZ = "rohde_schwarz"
    MINI_CIRCUITS = "mini_circuits"  # step attenuator
    KEITHLEY2600 = "keithley_2600"
    QBLOX_D5A = "D5a"
    QBLOX_S4G = "S4g"
    KEYSIGHT_E5080B = "keysight_e5080b"
    AGILENT_E5071B = "agilent_e5071B"
    YOKOGAWA_GS200 = "yokogawa_gs200"
    QCMRF = "QCM-RF"
    QUANTUM_MACHINES_CLUSTER = "quantum_machines_cluster"
    QDEVIL_QDAC2 = "qdevil_qdac2"


class VNAScatteringParameters(str, Enum):
    """Vector Network Analyzers Scattering Parameters
    Args:
        enum (str): Available types of scattering parameters:
        * S11
        * S12
        * S22
        * S21
    """

    S11 = "S11"
    S12 = "S12"
    S22 = "S22"
    S21 = "S21"


class VNATriggerModes(str, Enum):
    """Vector Network Analyzers Trigger Modes
    Args:
        enum (str): Available types of trigger modes:
        * INT
        * BUS
    """

    INT = "INT"
    BUS = "BUS"


class VNASweepModes(str, Enum):
    """Vector Network Analyzers Sweep Modes
    Args:
        enum (str): Available types of sweeping modes:
        * hold
        * cont
        * single
        * group
    """

    HOLD = "hold"
    CONT = "cont"
    SING = "single"
    GRO = "group"


class InstrumentControllerName(str, Enum):
    """Instrument Controller names.

    Args:
        enum (str): Available instrument controller element names:
        * qblox_pulsar
        * qblox_cluster
        * rohde_schwarz
        * mini_circuits
        * keithley_2600
        * keysight_e5080b
        * agilent_e5071B
        * yokogawa
        * qmm
    """

    QBLOX_CLUSTER = "qblox_cluster"
    ROHDE_SCHWARZ = "rohde_schwarz"
    MINI_CIRCUITS = "mini_circuits"  # step attenuator
    KEITHLEY2600 = "keithley_2600"
    QBLOX_SPIRACK = "qblox_spi_rack"
    KEYSIGHT_E5080B = "keysight_e5080b_controller"
    AGILENT_E5071B = "agilent_e5071B_controller"
    YOKOGAWA_GS200 = "yokogawa_gs200_controller"
    QUANTUM_MACHINES_CLUSTER = "quantum_machines_cluster_controller"
    QDEVIL_QDAC2 = "qdevil_qdac2"


@yaml.register_class
class Parameter(str, Enum):
    """Parameter names."""

    ALC = "alc"
    IQ_WIDEBAND = "iq_wideband"
    IQ_MODULATION = "iq_modulation"
    BUS_FREQUENCY = "bus_frequency"
    LO_FREQUENCY = "frequency"
    GAIN = "gain"
    DURATION = "duration"
    AMPLITUDE = "amplitude"
    PHASE = "phase"
    WAIT_TIME = "wait_time"
    DELAY_BETWEEN_PULSES = "delay_between_pulses"
    DELAY_BEFORE_READOUT = "delay_before_readout"
    GATE_DURATION = "gate_duration"
    GATE_PARAMETER = "gate_parameter"
    NUM_SIGMAS = "num_sigmas"
    DRAG_COEFFICIENT = "drag_coefficient"
    REFERENCE_CLOCK = "reference_clock"
    SEQUENCER = "sequencer"
    POWER = "power"
    GAIN_IMBALANCE = "gain_imbalance"
    PHASE_IMBALANCE = "phase_imbalance"
    SAMPLING_RATE = "sampling_rate"
    INTEGRATION = "integration"
    INTEGRATION_LENGTH = "integration_length"
    ATTENUATION = "attenuation"
    REPETITION_DURATION = "repetition_duration"
    SOFTWARE_AVERAGE = "software_average"
    SEQUENCE_TIMEOUT = "sequence_timeout"
    EXTERNAL = "external"
    RESET = "reset"
    HARDWARE_MODULATION = "hardware_modulation"
    HARDWARE_DEMODULATION = "hardware_demodulation"
    HARDWARE_INTEGRATION = "hardware_integration"
    SCOPE_ACQUIRE_TRIGGER_MODE = "scope_acquire_trigger_mode"
    SCOPE_HARDWARE_AVERAGING = "scope_hardware_averaging"
    IF = "intermediate_frequency"
    SOURCE_MODE = "source_mode"
    VOLTAGE = "voltage"
    CURRENT = "current"
    RAMPING_ENABLED = "ramping_enabled"
    RAMPING_RATE = "ramp_rate"
    SPAN = "span"
    LOW_PASS_FILTER = "low_pass_filter"  # noqa: S105
    SCATTERING_PARAMETER = "scattering_parameter"
    FREQUENCY_SPAN = "frequency_span"
    FREQUENCY_CENTER = "frequency_center"
    FREQUENCY_START = "frequency_start"
    FREQUENCY_STOP = "frequency_stop"
    IF_BANDWIDTH = "if_bandwidth"
    AVERAGING_ENABLED = "averaging_enabled"
    NUMBER_AVERAGES = "number_averages"
    TRIGGER_MODE = "trigger_mode"
    NUMBER_POINTS = "number_points"
    INTEGRATION_MODE = "integration_mode"
    ACQUISITION_TIMEOUT = "acquisition_timeout"
    MAX_CURRENT = "max_current"
    MAX_VOLTAGE = "max_voltage"
    SCOPE_STORE_ENABLED = "scope_store_enabled"
    TIME_OF_FLIGHT = "time_of_flight"
    SMEARING = "smearing"
    GAIN_I = "gain_i"
    GAIN_Q = "gain_q"
    OFFSET_I = "offset_i"
    OFFSET_Q = "offset_q"
    DC_OFFSET = "dc_offset"
    OFFSET_OUT0 = "offset_out0"
    OFFSET_OUT1 = "offset_out1"
    OFFSET_OUT2 = "offset_out2"
    OFFSET_OUT3 = "offset_out3"
    RF_ON = "rf_on"
    OPERATION_PARAMETER = "operation_parameter"
    DEVICE_TIMEOUT = "device_timeout"
    SWEEP_MODE = "sweep_mode"
    ELECTRICAL_DELAY = "electrical_delay"
    TIMEOUT = "timeout"
    NUM_FLIPS = "num_flips"
    OUTPUT_STATUS = "output_status"
    WEIGHTS_I = "weights_i"
    WEIGHTS_Q = "weights_q"
    WEIGHED_ACQ_ENABLED = "weighed_acq_enabled"
    THRESHOLD = "threshold"
    THRESHOLD_ROTATION = "threshold_rotation"
    OUT0_LO_FREQ = "out0_lo_freq"
    OUT0_IN0_LO_FREQ = "out0_in0_lo_freq"
    OUT1_LO_FREQ = "out1_lo_freq"
    OUT0_LO_EN = "out0_lo_en"
    OUT0_IN0_LO_EN = "out0_in0_lo_en"
    OUT0_IN0_LO_FREQ_CAL_TYPE_DEFAULT = "out0_in0_lo_freq_cal_type_default"
    OUT0_LO_FREQ_CAL_TYPE_DEFAULT = "out0_lo_freq_cal_type_default"
    OUT1_LO_FREQ_CAL_TYPE_DEFAULT = "out1_lo_freq_cal_type_default"
    OUT1_LO_EN = "out1_lo_en"
    OUT0_ATT = "out0_att"
    IN0_ATT = "in0_att"
    OUT1_ATT = "out1_att"
    OUT0_OFFSET_PATH0 = "out0_offset_path0"
    OUT1_OFFSET_PATH0 = "out1_offset_path0"
    OUT0_OFFSET_PATH1 = "out0_offset_path1"
    OUT1_OFFSET_PATH1 = "out1_offset_path1"
    DELAY = "delay"
    B = "b"
    T_PHI = "t_phi"
    GATE_OPTIONS = "options"
    FLUX = "flux"

    @classmethod
    def to_yaml(cls, representer, node):
        """Method to be called automatically during YAML serialization."""
        return representer.represent_scalar("!Parameter", f"{node.name}-{node.value}")

    @classmethod
    def from_yaml(cls, _, node):
        """Method to be called automatically during YAML deserialization."""
        _, value = node.value.split("-")
        return cls(value)


class ResultName(str, Enum):
    """Result names.

    Args:
        enum (str): Available result element names:
        * qblox
    """

    QBLOX = "qblox"
    QBLOX_QPROGRAM_MEASUREMENT = "qblox_qprogram_measurement"
    VECTOR_NETWORK_ANALYZER = "vector_network_analyzer"
    QUANTUM_MACHINES = "quantum_machines"
    QUANTUM_MACHINES_MEASUREMENT = "quantum_machines_measurement"


class ConnectionName(str, Enum):
    """Connection names.

    Args:
        enum (str): Available connection element names:
        * tcp_ip
        * usb
    """

    TCP_IP = "tcp_ip"
    USB = "usb"


class InstrumentTypeName(str, Enum):
    """Instrument Type names (the name of the class).

    Args:
        enum (str): Available instrument type element names:
        * QbloxQCM
        * QbloxQRM
        * SGS100A
        * Attenuator
        * Keithley2600
        * QbloxD5a
        * QbloxS4g
        * YokogawaGS200
        * QDevilQDac2
    """

    QBLOX_QCM = "QbloxQCM"
    QBLOX_QRM = "QbloxQRM"
    ROHDE_SCHWARZ = "SGS100A"
    MINI_CIRCUITS = "Attenuator"
    KEITHLEY2600 = "Keithley2600"
    QBLOX_D5A = "QbloxD5a"
    QBLOX_S4G = "QbloxS4g"
    YOKOGAWA_GS200 = "YokogawaGS200"
    QDEVIL_QDAC2 = "QDevilQDac2"


class ResetMethod(str, Enum):
    PASSIVE = "passive"
    ACTIVE = "active"


class SourceMode(str, Enum):
    """Source Modes"""

    CURRENT = "current"
    VOLTAGE = "voltage"


class Line(str, Enum):
    """Chip line"""

    FLUX = "flux"
    DRIVE = "drive"
    READOUT = "readout"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Vector Network Analyzer generic PyVisa driver"""
from dataclasses import dataclass, field

import pyvisa

from qililab.constants import DEFAULT_TIMEOUT
from qililab.typings.instruments.device import Device


@dataclass
class VectorNetworkAnalyzerDriver(Device):
    """Typing class of the driver for a Vector Network Analyzer instrument."""

    name: str
    address: str
    driver: pyvisa.Resource = field(init=False)
    timeout: float = DEFAULT_TIMEOUT

    def __post_init__(self):
        """Configure driver and connect to the resource"""
        resource_manager = pyvisa.ResourceManager("@py")
        self.driver = resource_manager.open_resource(f"TCPIP::{self.address}::INSTR")
        self.driver.timeout = self.timeout

    def initial_setup(self):
        """Set initial preset"""
        self.driver.write("FORM:DATA REAL,32")
        self.driver.write("*CLS")
        self.reset()

    def reset(self):
        """Reset instrument settings."""
        self.driver.write("SYST:PRES; *OPC?")

    def send_command(self, command: str, arg: str = "?"):
        """Function to communicate with the device."""
        return self.driver.write(f"{command} {arg}")  # type: ignore

    def send_query(self, query: str):
        """Function to communicate with the device."""
        return self.driver.query(query)  # type: ignore

    def send_binary_query(self, query: str):
        """Function to communicate with the device."""
        return self.driver.query_binary_values(query)  # type: ignore

    def set_timeout(self, value: float):
        """Set timeout in mili seconds."""
        self.timeout = value
        self.driver.timeout = self.timeout

    def read(self) -> str:
        """read directly from the device"""
        return self.driver.read()  # type: ignore

    def read_raw(self):
        """read raw data directly from the device"""
        return self.driver.read_raw()  # type: ignore



"""Tests for the PulseShape class."""
import itertools

import numpy as np
import pytest

from qililab.pulse.pulse_shape import PulseShape, Rectangular

from .helper_functions import return_envelope

# Parameters of the envelope.
DURATION = [50, 25, 500]
AMPLITUDE = [0, 0.9, -1.0, 1.2]
RESOLUTION = [1.0, 0.1]


@pytest.fixture(
    name="env_params",
    params=[
        {"duration": duration, "amplitude": amplitude, "resolution": resolution}
        for duration, amplitude, resolution in itertools.product(DURATION, AMPLITUDE, RESOLUTION)
    ],
)
def fixture_env_params(request: pytest.FixtureRequest) -> list:
    """Fixture for the envelope parameters."""
    return request.param


@pytest.mark.parametrize("pulse_shape", [Rectangular()])
class TestPulseShape:
    """Unit tests checking the PulseShape attributes and methods"""

    def test_envelope_methods(self, pulse_shape: Rectangular, env_params: dict[str, int]):
        """Test the envelope method"""
        env = return_envelope(pulse_shape, env_params)

        # Test not None and type
        assert env is not None
        assert isinstance(env, np.ndarray)

        # Assert size of np.ndarray
        assert len(env) == env_params["duration"] / env_params["resolution"]

    def test_max_min_of_envelope_method(self, pulse_shape: Rectangular, env_params: dict[str, int]):
        """Test the maximums and minimums of the envelope method"""
        env = return_envelope(pulse_shape, env_params)

        # Test  the maximums of the positive envelopes
        if env_params["amplitude"] > 0:
            assert round(np.max(np.real(env)), 2) == np.max(env_params["amplitude"])

        # Test the minimums of the negative envelopes
        elif env_params["amplitude"] < 0:
            assert round(np.min(np.real(env)), 2) == np.min(env_params["amplitude"])

        # Test the 0 amplitude case
        elif env_params["amplitude"] == 0:
            assert np.min(np.real(env)) == np.max(np.real(env)) == 0

    def test_envelope_method_shapes(self, pulse_shape: Rectangular, env_params: dict[str, int]):
        """Test shapes of the envelopes"""
        env = return_envelope(pulse_shape, env_params)
        assert np.max(env) == np.min(env)

    def test_from_dict(self, pulse_shape: Rectangular):
        """Test for the to_dict method."""
        dictionary = pulse_shape.to_dict()
        pulse_shape2: Rectangular = Rectangular.from_dict(dictionary)

        dictionary2 = pulse_shape2.to_dict()
        pulse_shape3: Rectangular = Rectangular.from_dict(dictionary2)

        for shape in [pulse_shape2, pulse_shape3]:
            assert shape is not None
            assert isinstance(shape, PulseShape)
            assert isinstance(shape, Rectangular)

        assert pulse_shape == pulse_shape2 == pulse_shape3

    def test_to_dict_method(self, pulse_shape: Rectangular):
        """Test to_dict method"""
        dictionary = pulse_shape.to_dict()

        pulse_shape2: Rectangular = Rectangular.from_dict(dictionary)
        dictionary2 = pulse_shape2.to_dict()

        for dict_ in [dictionary, dictionary2]:
            assert dict_ is not None
            assert isinstance(dict_, dict)

        assert (
            dictionary
            == dictionary2
            == {
                "name": pulse_shape.name.value,
            }
        )

    def test_envelope_with_amplitude_0(self, pulse_shape: Rectangular):
        """Testing that the corner case amplitude = 0 works properly."""
        ENV_DURATION = 10
        envelope = pulse_shape.envelope(amplitude=0, duration=ENV_DURATION)
        assert np.allclose(envelope, np.zeros(ENV_DURATION))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QDevil QDAC-II Controller"""
from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentControllerSettings
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.qdevil.qdevil_qdac2 import QDevilQDac2
from qililab.typings import QDevilQDac2 as QDevilQDac2Device
from qililab.typings.enums import ConnectionName, InstrumentControllerName, InstrumentTypeName


@InstrumentControllerFactory.register
class QDevilQDac2Controller(SingleInstrumentController):
    """QDevil QDAC-II controller class
    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        device (QDevilQDac2Device): Instance of the QDevil QDAC-II class.
        settings (QDevilQDac2ControllerSettings): Settings of the instrument.
    """

    name = InstrumentControllerName.QDEVIL_QDAC2

    @dataclass
    class QDevilQDac2ControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific GS200 Controller."""

    settings: QDevilQDac2ControllerSettings
    device: QDevilQDac2Device
    modules: Sequence[QDevilQDac2]

    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""
        if self.settings.connection.name == ConnectionName.TCP_IP:
            self.device = QDevilQDac2Device(f"{self.name.value}", f"TCPIP::{self.address}::5025::SOCKET")
        else:
            self.device = QDevilQDac2Device(f"{self.name.value}", f"ASRL/dev/{self.address}::INSTR")

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, QDevilQDac2):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument is {InstrumentTypeName.QDEVIL_QDAC2}"
                )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class used as hashtable to load the class corresponding to a given category"""

from enum import Enum
from typing import ClassVar, TypeVar

from qililab.typings.factory_element import FactoryElement

Element = TypeVar("Element", bound=FactoryElement)


class Factory:
    """Hash table that loads a specific class given an object's name."""

    handlers: ClassVar[dict[str, type[FactoryElement]]] = {}

    @classmethod
    def register(cls, handler_cls: type[Element]):
        """Register handler in the factory.

        Args:
            output_type (type): Class type to register.
        """
        cls.handlers[handler_cls.name.value] = handler_cls
        return handler_cls

    @classmethod
    def get(cls, name: str | Enum):
        """Return class attribute."""
        return cls.handlers[name.value] if isinstance(name, Enum) else cls.handlers[name]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrument Reference class."""

from dataclasses import dataclass


@dataclass
class InstrumentReference:
    """References an Instrument with its alias to be retrieved from the Instrument Factory.

    Args:
        alias (str): The alias name of the Instrument.
        slot_id (int): The number that identifies the slot used by the instrument within the
                            the possible list of available modules (i.e. on a Cluster).
    """

    alias: str
    slot_id: int  # slot_id represents the number displayed in the cluster

    def __iter__(self):
        """Iterate over InstrumentReference elements.

        Yields:
            tuple[tuple[str], tuple[str, int], tuple[str, int]]: alias and slot_id
        """
        yield from self.__dict__.items()

    def to_dict(self):
        """Return a dict representation of the InstrumentReference class."""
        return {"alias": self.alias, "slot_id": self.slot_id}

    @classmethod
    def from_dict(cls, settings: dict):
        """Build an InstrumentReference from a settings dictionary

        Args:
            settings (dict): an instrument reference from the settings file
        """
        return cls(**settings)
